/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cassert>
#include <cstring>
#include <memory>
#include <string>
#include <vector>
#include "defs.h"
#include "record/rm_defs.h"
#include <map>
#include <ctime>
#include <regex>

struct TabCol {
    std::string tab_name;
    std::string col_name;
    //AgreType agre_type = AgreType::NONE;
    friend bool operator<(const TabCol &x, const TabCol &y) {
        return std::make_pair(x.tab_name, x.col_name) < std::make_pair(y.tab_name, y.col_name);
    }
};

struct AgreCol {
    TabCol alias;//投影的函数名,化名
    TabCol agre_col;//聚合的目标列
    AgreType agre_type;//聚合类型
};

struct Value {
    ColType type;  // type of value
    union {
        int int_val;      // int value
        float float_val;  // float value
        time_t date_val;  //date value
    };
    std::string str_val;  // string value
    std::shared_ptr<RmRecord> raw;  // raw record buffer


    std::unordered_set<int> int_set;
    std::unordered_set<float> float_set;
    std::unordered_set<std::string> string_set;

    void set_int(int int_val_) {
        type = TYPE_INT;
        int_val = int_val_;
    }

    void set_float(float float_val_) {
        type = TYPE_FLOAT;
        float_val = float_val_;
    }

    void set_str(std::string str_val_) {
        type = TYPE_STRING;
        str_val = std::move(str_val_);
    }


    void add_to_set(Value &tmp){
        //assert(is_set);
        switch (tmp.type) {
            case TYPE_INT:
                set_int(tmp.int_val);
                int_set.insert(tmp.int_val);
                break;
            case TYPE_FLOAT:
                set_float(tmp.float_val);
                float_set.insert(tmp.float_val);
                break;
            case TYPE_STRING:
                set_str(tmp.str_val);
                string_set.insert(tmp.str_val);
                break;

            default:
                // Handle error or unexpected type
                break;
        }
    }

    void init_raw(int len) {
        assert(raw == nullptr);
        //if(bug_flag)throw DeBugError();
        raw = std::make_shared<RmRecord>(len);
        if (type == TYPE_INT) {
            assert(len == sizeof(int));
            *(int *)(raw->data) = int_val;
        } else if (type == TYPE_FLOAT) {
            assert(len == sizeof(float));
            *(float *)(raw->data) = float_val;
        } else if (type == TYPE_STRING) {
            if (len < (int)str_val.size()) {
                //assert(0);
                throw StringOverflowError();
            }
            memset(raw->data, 0, len);
            memcpy(raw->data, str_val.c_str(), str_val.size());
        }else if(type == TYPE_DATETIME){
            memset(raw->data, 0, len);
            memcpy(raw->data, str_val.c_str(), str_val.size());
        }
    }

    void fill(int len, char* buffer) {
        switch (type) {
            case TYPE_INT:{
                assert(len == sizeof(int));
                *(int *)(buffer) = int_val;
                break;
            }
            case TYPE_FLOAT:{
                assert(len == sizeof(float));
                *(float *)(buffer) = float_val;
                break;
            }
            case TYPE_STRING:{
                if (len < (int)str_val.size()) {
                    throw StringOverflowError();
                }
                memset(buffer, 0, len);
                memcpy(buffer, str_val.c_str(), str_val.size());
                break;
            }
            default:
                assert(0);
        }
    }


    bool is_in_set(const Value& other) const {
        if(other.type != this->type)throw IncompatibleOpError();
        switch (this->type) {
            case TYPE_INT:
                return int_set.find(other.int_val) != int_set.end();
            case TYPE_FLOAT:
                return float_set.find(other.float_val) != float_set.end();
            case TYPE_STRING:
                return string_set.find(other.str_val) != string_set.end();
            default:
                assert(0);
        }
    }

    // 重载value的比较运算符 >,<,=,!
    bool operator<(const Value& right_value) const {
        switch (type) {
            case TYPE_INT:
                //assert(right_value.int_val != std::numeric_limits<int>::min() && right_value.int_val != std::numeric_limits<int>::max());
                if(right_value.type == TYPE_INT) {
                    return int_val < right_value.int_val;
                }else if(right_value.type == TYPE_FLOAT) {
                    return int_val < right_value.float_val;
                }else{
                    throw IncompatibleTypeError(coltype2str(ColType::TYPE_INT), coltype2str(right_value.type));
                }
            case TYPE_FLOAT:
                //assert(right_value.float_val != std::numeric_limits<float>::lowest() && right_value.float_val != std::numeric_limits<float>::max());
                if (right_value.type == TYPE_INT) {
                    return float_val < right_value.int_val;
                } else if (right_value.type == TYPE_FLOAT) {
                    return float_val < right_value.float_val;
                }else{
                    throw IncompatibleTypeError(coltype2str(ColType::TYPE_INT), coltype2str(right_value.type));
                }
            case TYPE_STRING:
            case TYPE_DATETIME:
                if(right_value.type == TYPE_INT || right_value.type == TYPE_FLOAT){
                    throw IncompatibleTypeError(coltype2str(ColType::TYPE_STRING), coltype2str(right_value.type));
                }
                return (strcmp(str_val.c_str(), right_value.str_val.c_str()) < 0 );
            default:
                assert(0);
        }
        throw IncompatibleTypeError(coltype2str(type), coltype2str(right_value.type));
    }

    bool operator>(const Value& right_value) const {
        switch (type) {
            case TYPE_INT:
                //assert(right_value.int_val != std::numeric_limits<int>::min() && right_value.int_val != std::numeric_limits<int>::max());
                if(right_value.type == TYPE_INT) {
                    return int_val > right_value.int_val;
                }else if(right_value.type == TYPE_FLOAT) {
                    return int_val > right_value.float_val;
                }else{
                    throw IncompatibleTypeError(coltype2str(ColType::TYPE_INT), coltype2str(right_value.type));
                }
            case TYPE_FLOAT:
                if (right_value.type == TYPE_INT) {
                    return float_val > right_value.int_val;
                } else if (right_value.type == TYPE_FLOAT) {
                    return float_val > right_value.float_val;
                }else{
                    throw IncompatibleTypeError(coltype2str(ColType::TYPE_INT), coltype2str(right_value.type));
                }
            case TYPE_STRING:
            case TYPE_DATETIME:
                if(right_value.type == TYPE_INT || right_value.type == TYPE_FLOAT){
                    throw IncompatibleTypeError(coltype2str(ColType::TYPE_STRING), coltype2str(right_value.type));
                }
                return (strcmp(str_val.c_str(), right_value.str_val.c_str()) > 0 );
            default:
                assert(0);
        }
        throw IncompatibleTypeError(coltype2str(type), coltype2str(right_value.type));
    }

    bool operator==(const Value& right_value) const {
        switch (type) {
            case TYPE_INT:
                if(right_value.type == TYPE_INT) {
                    return int_val == right_value.int_val;
                }else if(right_value.type == TYPE_FLOAT) {
                    return int_val == right_value.float_val;
                }else{
                    throw IncompatibleTypeError(coltype2str(ColType::TYPE_INT), coltype2str(right_value.type));
                }
            case TYPE_FLOAT:
                if (right_value.type == TYPE_INT) {
                    return float_val == right_value.int_val;
                } else if (right_value.type == TYPE_FLOAT) {
                    return float_val == right_value.float_val;
                }else{
                    throw IncompatibleTypeError(coltype2str(ColType::TYPE_INT), coltype2str(right_value.type));
                }
            case TYPE_STRING:
            case TYPE_DATETIME:
                if(right_value.type == TYPE_INT || right_value.type == TYPE_FLOAT){
                    throw IncompatibleTypeError(coltype2str(ColType::TYPE_STRING), coltype2str(right_value.type));
                }
                return (strcmp(str_val.c_str(), right_value.str_val.c_str()) == 0 );
            default:
                assert(0);
        }
        throw IncompatibleTypeError(coltype2str(type), coltype2str(right_value.type));
    }
    bool operator!=(const Value& right_value) const {
        return !(*this == right_value);
    }
};

enum CompOp { OP_EQ, OP_NE, OP_LT, OP_GT, OP_LE, OP_GE, OP_IN,OP_NONE };

struct AgreValue {
    AgreType agre_type;
    ColType type;
    union {
        int int_val;            // int value
        float float_val;        // float value
    };
    char* str_val;  // string value
    int str_len;

    //针对不同类型字段，所采用的map
    std::map<int,int> int_map;
    std::map<float,int> float_map;
    std::map<std::string,int> str_map;
//    union{
//        int count;
//        //int all_count;
//    };
    int count = 0;
    void init_agre_raw(ColType value_type, AgreType agre_type_, int len) {
        type = value_type;
        agre_type = agre_type_;
        if(type != TYPE_INT && type != TYPE_FLOAT && type != TYPE_STRING) {
            assert(0);
        }
        if(type == TYPE_STRING) {
            str_len = len;
        }
        if (agre_type == AGRE_TYPE_MAX) {
            if (type == TYPE_INT) {
                assert(len == sizeof(int));
                int_val = std::numeric_limits<int>::min();
            } else if (type == TYPE_FLOAT) {
                assert(len == sizeof(float));
                float_val = std::numeric_limits<float>::lowest();
            } else if (type == TYPE_STRING) {
                str_val = new char[len];
                memset(str_val, 0, len);
            }
        } else if (agre_type == AGRE_TYPE_MIN) {
            if (type == TYPE_INT) {
                assert(len == sizeof(int));
                int_val = std::numeric_limits<int>::max();
            } else if (type == TYPE_FLOAT) {
                assert(len == sizeof(float));
                float_val = std::numeric_limits<float>::max();
            } else if (type == TYPE_STRING) {
                str_val = new char[len];
                memset(str_val, 0xff, len);
            }
        } else if (agre_type == AGRE_TYPE_SUM) {
            if (type == TYPE_INT) {
                assert(len == sizeof(int));
                int_val = 0;
            } else if (type == TYPE_FLOAT) {
                assert(len == sizeof(float));
                float_val = 0;
            }
            else if(type == TYPE_STRING) {
                assert(0);
            }
        }else if(agre_type == AGRE_TYPE_COUNT){
            int_val = 0;
        }else if(agre_type == AGRE_TYPE_COUNT_ALL){
            int_val = 0;
        }else{
            assert(0);
        }
    }

    void add_value(const char* value, int len) {
        switch (agre_type) {
            case AGRE_TYPE_COUNT_ALL: {
                int_val += 1;
                break;
            }

            case AGRE_TYPE_MAX:
                switch (type) {
                    case TYPE_INT: {
                        //assert(len == sizeof(int));
                        int value_ = *(int *)(value);
                        if (value_ > int_val) {
                            int_val = value_;
                        }
                        break;
                    }
                    case TYPE_FLOAT: {
                        //assert(len == sizeof(float));
                        float value_ = *(float *)(value);
                        if (value_ > float_val) {
                            float_val = value_;
                        }
                        break;
                    }
                    case TYPE_STRING: {
                        if (memcmp(value, str_val, len) > 0) {
                            memcpy(str_val, value, len);
                        }
                        break;
                    }
                    default:
                        assert(0);
                }
                break;

            case AGRE_TYPE_MIN:
                switch (type) {
                    case TYPE_INT: {
                        //assert(len == sizeof(int));
                        int value_ = *(int *)(value);
                        if (value_ < int_val) {
                            int_val = value_;
                        }
                        break;
                    }
                    case TYPE_FLOAT: {
                        //assert(len == sizeof(float));
                        float value_ = *(float *)(value);
                        if (value_ < float_val) {
                            float_val = value_;
                        }
                        break;
                    }
                    case TYPE_STRING: {
                        if (memcmp(value, str_val, len) < 0) {
                            memcpy(str_val, value, len);
                        }
                        break;
                    }
                    default:
                        assert(0);
                }
                break;

            case AGRE_TYPE_SUM:
                switch (type) {
                    case TYPE_INT: {
                        int value_ = *(int *)(value);
                        int_val += value_;
                        break;
                    }
                    case TYPE_FLOAT: {
                        float value_ = *(float *)(value);
                        float_val += value_;
                        break;
                    }
                    default:
                        assert(0);
                }
                break;

            case AGRE_TYPE_COUNT:{
                //int_val += 1;
                switch (type) {
                    case TYPE_INT: {
                        //assert(len == sizeof(int));
                        int value_ = *(int *)(value);
                        int_map[value_] = 0;
                        int_val += 1;
                        break;
                    }
                    case TYPE_FLOAT: {
                        //assert(len == sizeof(float));
                        float value_ = *(float *)(value);
                        float_map[value_] = 0;
                        int_val += 1;
                        break;
                    }
                    case TYPE_STRING: {
                        //assert(len == str_len);
                        std::string value_(value, len);
                        str_map[value_] = 0;
                        int_val += 1;
                        break;
                    }
                    default:
                        assert(0);
                }
                break;
            }
            default:
                assert(0);
        }
    }

    int get_len(){
        if(agre_type == AGRE_TYPE_COUNT_ALL || agre_type == AGRE_TYPE_COUNT){
            return sizeof(int);
        }
        switch (type) {
            case ColType::TYPE_INT:
                return sizeof(int);
            case ColType::TYPE_FLOAT:
                return sizeof(float);
            case ColType::TYPE_STRING:
                return str_len;
            default:
                assert(0);
        }
        return -1;
    }
    std::unique_ptr<RmRecord> get_rec() {
        if (agre_type == AGRE_TYPE_MAX || agre_type == AGRE_TYPE_MIN || agre_type == AGRE_TYPE_SUM) {
            if (type == TYPE_INT) {
                std::unique_ptr<RmRecord> ret = std::make_unique<RmRecord>((int)sizeof(int));
                *(int *)ret->data = int_val;
                return ret;
            } else if (type == TYPE_FLOAT) {
                std::unique_ptr<RmRecord> ret = std::make_unique<RmRecord>((int)sizeof(float));
                *(float *)ret->data = float_val;
                return ret;
            } else if (type == TYPE_STRING) {
                if(agre_type == AGRE_TYPE_SUM) {
                    assert(0);
                }
                else {
                    std::unique_ptr<RmRecord> ret = std::make_unique<RmRecord>(str_len);
                    memcpy(ret->data, str_val, str_len);
                    return ret;
                }
            }
        } else if (agre_type == AGRE_TYPE_COUNT || agre_type == AGRE_TYPE_COUNT_ALL) {
            std::unique_ptr<RmRecord> ret = std::make_unique<RmRecord>((int)sizeof(int));
            *(int *)ret->data = int_val;
            return ret;
        }

        return nullptr;
    }

    Value to_value() const {
        Value value;
        if (agre_type == AGRE_TYPE_COUNT || agre_type == AGRE_TYPE_COUNT_ALL){
            value.set_int(int_val);
            return value;
        }
        value.type = type;
        switch (type) {
            case TYPE_INT:
                value.set_int(int_val);
                break;
            case TYPE_FLOAT:
                value.set_float(float_val);
                break;
            case TYPE_STRING:
                value.set_str(std::string(str_val, str_len));
                break;
            default:
                assert(0);//未
        }
        return value;
    }

};

class Query;
class Plan;
class AbstractExecutor;

struct SubQueryInfo{
    std::shared_ptr<Query> sub_query;
    std::shared_ptr<Plan>  sub_query_plan;
    std::shared_ptr<AbstractExecutor> sub_query_exec;
};

struct Condition {
    TabCol lhs_col;   // left-hand side column
    CompOp op;        // comparison operator
    bool is_rhs_val;  // true if right-hand side is a value (not a column)
    bool is_rhs_sub_query; // 右值是否为无关子查询，后续可以改为union，重构代码
    TabCol rhs_col;   // right-hand side column
    Value rhs_val;    // right-hand side value
    std::shared_ptr<SubQueryInfo> sub_query_info;
};



struct AgreCondition {
    AgreCol agre_lhs_col;
    CompOp op;        // comparison operator
    Value rhs_val;    // right-hand side value
};

enum UpdateOp
{
    NO = 0, PLUS, MINUS, MULTI, DIVIDE
};

struct SetClause {
    TabCol lhs;
    Value rhs;
    UpdateOp op;
};

namespace std {
    template <>
    struct hash<Value> {
        std::size_t operator()(const Value& v) const noexcept {
            std::size_t h = 0;
            switch (v.type) {
                case TYPE_INT:
                    h = std::hash<int>()(v.int_val);
                    break;
                case TYPE_FLOAT:
                    h = std::hash<float>()(v.float_val);
                    break;
                case TYPE_STRING:
                    h = std::hash<std::string>()(v.str_val);
                    break;
                default:
                    assert(0);
            }
            return h;
        }
    };
}

