/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class DeleteExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;                   // 表的元数据
    std::vector<Condition> conds_;  // delete的条件
    RmFileHandle *fh_;              // 表的数据文件句柄
    std::vector<Rid> rids_;         // 需要删除的记录的位置
    std::string tab_name_;          // 表名称
    SmManager *sm_manager_;
    bool isend;

   public:
    DeleteExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<Condition> conds,
                   std::vector<Rid> rids, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = std::move(conds);
        rids_ = std::move(rids);
        context_ = context;
        isend = false;
        if (!context_->lock_mgr_->lock_IX_on_table(context_->txn_, fh_->GetFd())) {
            throw TransactionAbortException(context_->txn_->get_transaction_id(),AbortReason::UPGRADE_CONFLICT);
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        // 遍历所有的记录
        for (auto &rid : rids_) {
            auto record = fh_->get_record(rid, context_);
            if (enable_logging) {
                //事务提交，将写操作写进txn对应的写集合
                auto delete_log = new DeleteLogRecord(
                        context_->txn_->get_transaction_id(),
                        context_->txn_->get_prev_lsn(),
                        *record,
                        rid,
                        tab_name_
                );
                context_->log_mgr_->add_log_to_buffer(delete_log);
                context_->txn_->set_prev_lsn(delete_log->lsn_);
                Page *page = sm_manager_->get_bpm()->fetch_page({fh_->GetFd(), rid.page_no});
                page->set_page_lsn(delete_log->lsn_);
                sm_manager_->get_bpm()->unpin_page({fh_->GetFd(), rid.page_no}, true);
            }
            auto write_record = new WriteRecord(WType::DELETE_TUPLE, tab_name_, rid, *record);
            context_->txn_->append_write_record(write_record);

            fh_->delete_record(rid, context_);
            // delete in index
            for (size_t i = 0; i < tab_.indexes.size(); ++i) {
                auto &index = tab_.indexes[i];
                auto ih = sm_manager_->ihs_.at(
                        sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
                char *key = new char[index.col_tot_len];
                int offset = 0;
                for (size_t j = 0; j < (size_t)index.col_num; ++j) {
                    memcpy(key + offset, write_record->GetRecord().data + index.cols[j].offset, index.cols[j].len);
                    offset += index.cols[j].len;
                }
                //间隙锁获取
//                if(context_->txn_->get_isolation_level() == IsolationLevel::SERIALIZABLE){
//                    Rid gap_rid;
//                    ih->upper_bound(key,gap_rid);
//                    //context_->lock_mgr_->try_lock_gap_on_index(context_->txn_,ih->get_fd(),gap_rid);
//                    context_->lock_mgr_->try_erase_gap_on_index(context_->txn_,ih->get_fd(),gap_rid);//删除同一事务的gap
//                }
                ih->delete_entry(key, context_->txn_);
                delete[] key;
            }
        }
        isend = true;
        return nullptr;
    }

    std::string getType() override { return "DeleteExecutor"; }

    bool is_end() const override { return isend; }

    Rid &rid() override { return _abstract_rid; }

    ColMeta get_col_offset(const TabCol &target) override {
        ColMeta result;
        for (const auto &cond : conds_) {
            if (cond.lhs_col.col_name == target.col_name && cond.lhs_col.tab_name == target.tab_name) {
                result = *(tab_.get_col(cond.lhs_col.col_name));
            }
        }
        return result;
    }
};