/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include <cmath>
#include "ix_index_handle.h"

#include "ix_scan.h"
/**
 * @brief 在当前node中查找第一个>=target的key_idx
 *
 * @return key_idx，范围为[0,num_key)，如果返回的key_idx=num_key，则表示target大于最后一个key
 * @note 返回key index（同时也是rid index），作为slot no
 */
int IxNodeHandle::lower_bound(const char *target) const {
    // Todo:
    // 查找当前节点中第一个大于等于target的key，并返回key的位置给上层
    // 提示: 可以采用多种查找方式，如顺序遍历、二分查找等；使用ix_compare()函数进行比较

    // 二分查找
    int left = 0;
    int right = page_hdr->num_key;
    while (left < right) {
        int mid = left + (right - left) / 2;
        int cmp = ix_compare(get_key(mid), target, file_hdr->col_types_, file_hdr->col_lens_);
        if (cmp < 0) {
            left = mid + 1;
        } else {
            right = mid;
        }
    }
    return left;
}

/**
 * @brief 在当前node中查找第一个>target的key_idx
 *
 * @return key_idx，范围为[1,num_key)，如果返回的key_idx=num_key，则表示target大于等于最后一个key
 * @note 注意此处的范围从1开始
 */
int IxNodeHandle::upper_bound(const char *target) const {
    // Todo:
    // 查找当前节点中第一个大于target的key，并返回key的位置给上层
    // 提示: 可以采用多种查找方式：顺序遍历、二分查找等；使用ix_compare()函数进行比较

    // 二分查找
    int left = 1;
    int right = page_hdr->num_key;
    while (left < right) {
        int mid = left + (right - left) / 2;
        int cmp = ix_compare(get_key(mid), target, file_hdr->col_types_, file_hdr->col_lens_);

        if (cmp <= 0) {
            left = mid + 1;
        } else {
            right = mid;
        }
    }
    return left;
}

/**
 * @brief 用于叶子结点根据key来查找该结点中的键值对
 * 值value作为传出参数，函数返回是否查找成功
 *
 * @param key 目标key
 * @param[out] value 传出参数，目标key对应的Rid
 * @return 目标key是否存在
 */
bool IxNodeHandle::leaf_lookup(const char *key, Rid **value) {
    // Todo:
    // 1. 在叶子节点中获取目标key所在位置
    // 2. 判断目标key是否存在
    // 3. 如果存在，获取key对应的Rid，并赋值给传出参数value
    // 提示：可以调用lower_bound()和get_rid()函数。

    int index = lower_bound(key);
    if(index >= page_hdr->num_key || ix_compare(get_key(index), key, file_hdr->col_types_, file_hdr->col_lens_) != 0){
        return false;
    }
    *value = get_rid(index);
    return true;
}

/**
 * 用于内部结点（非叶子节点）查找目标key所在的孩子结点（子树）
 * @param key 目标key
 * @return page_id_t 目标key所在的孩子节点（子树）的存储页面编号
 */
page_id_t IxNodeHandle::internal_lookup(const char *key) {
    // Todo:
    // 1. 查找当前非叶子节点中目标key所在孩子节点（子树）的位置
    // 2. 获取该孩子节点（子树）所在页面的编号
    // 3. 返回页面编号

    int index = upper_bound(key); //找到第一个大于target key的key
    return get_rid(index - 1)->page_no;
}

/**
 * @brief 在指定位置插入n个连续的键值对
 * 将key的前n位插入到原来keys中的pos位置；将rid的前n位插入到原来rids中的pos位置
 *
 * @param pos 要插入键值对的位置
 * @param (key, rid) 连续键值对的起始地址，也就是第一个键值对，可以通过(key, rid)来获取n个键值对
 * @param n 键值对数量
 * @note [0,pos)           [pos,num_key)
 *                            key_slot
 *                            /      \
 *                           /        \
 *       [0,pos)     [pos,pos+n)   [pos+n,num_key+n)
 *                      key           key_slot
 */
void IxNodeHandle::insert_pairs(int pos, const char *key, const Rid *rid, int n) {
    // Todo:
    // 1. 判断pos的合法性
    // 2. 通过key获取n个连续键值对的key值，并把n个key值插入到pos位置
    // 3. 通过rid获取n个连续键值对的rid值，并把n个rid值插入到pos位置
    // 4. 更新当前节点的键数量


//    if(pos < 0 || page_hdr->num_key + n > get_max_size()) {
//        return;
//    }
    if(pos < 0 || page_hdr->num_key + n - 1 > file_hdr->btree_order_) {
        return;
    }
    for (int i = page_hdr->num_key - pos; i > 0; i--) {
        set_key(pos + i + n - 1 , get_key(pos + i - 1));
        set_rid(pos + i + n - 1, rids[pos + i - 1]);
    }
    for(int i = 0; i < n; i++){
        set_key(pos + i, key + i * file_hdr->col_tot_len_ );
        set_rid(pos + i, rid[i]);
    }
    page_hdr->num_key += n;
}

/**
 * @brief 用于在结点中插入单个键值对。
 * 函数返回插入后的键值对数量
 *
 * @param (key, value) 要插入的键值对
 * @return int 键值对数量
 */
int IxNodeHandle::insert(const char *key, const Rid &value) {
    // Todo:
    // 1. 查找要插入的键值对应该插入到当前节点的哪个位置
    // 2. 如果key重复则不插入
    // 3. 如果key不重复则插入键值对
    // 4. 返回完成插入操作之后的键值对数量

    int index = lower_bound(key);
    if(index < page_hdr->num_key && ix_compare(get_key(index), key, file_hdr->col_types_, file_hdr->col_lens_) == 0){
        // 重复，不插入
        return page_hdr->num_key;
    }
    insert_pair(index, key, value);
    return page_hdr->num_key;
}

/**
 * @brief 用于在结点中的指定位置删除单个键值对
 *
 * @param pos 要删除键值对的位置
 */
void IxNodeHandle::erase_pair(int pos) {
    // Todo:
    // 1. 删除该位置的key
    // 2. 删除该位置的rid
    // 3. 更新结点的键值对数量

    for(int i = pos; i < get_size(); i++){
        set_key(i, get_key(i + 1));
        set_rid(i, *get_rid(i + 1));
    }
    set_size(get_size() - 1);
    return;
}

/**
 * @brief 用于在结点中删除指定key的键值对。函数返回删除后的键值对数量
 *
 * @param key 要删除的键值对key值
 * @return 完成删除操作后的键值对数量
 */
int IxNodeHandle::remove(const char *key) {
    // Todo:
    // 1. 查找要删除键值对的位置
    // 2. 如果要删除的键值对存在，删除键值对
    // 3. 返回完成删除操作后的键值对数量

    int index = lower_bound(key);
    if(index < page_hdr->num_key && ix_compare(get_key(index), key, file_hdr->col_types_, file_hdr->col_lens_) == 0){
        erase_pair(index);
    }
    return get_size();
}

/**
 * @brief 判断节点是否可以安全放锁
*/
bool IxNodeHandle::is_safe(Operation op) {
    int min_sz = is_root_page() ? 2 : get_min_size();
    switch (op)
    {
    case Operation::FIND:
        return true;
        break;
    case Operation::INSERT:
        return get_size() + 1 < get_max_size();
        break;
    case Operation::DELETE:
        return get_size() - 1 >= min_sz;
        break;
    default:
        break;
    }
    return false;
}

IxIndexHandle::IxIndexHandle(DiskManager *disk_manager, BufferPoolManager *buffer_pool_manager, int fd)
        : disk_manager_(disk_manager), buffer_pool_manager_(buffer_pool_manager), fd_(fd) {
    // init file_hdr_
    disk_manager_->read_page(fd, IX_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
    char* buf = new char[PAGE_SIZE];
    memset(buf, 0, PAGE_SIZE);
    disk_manager_->read_page(fd, IX_FILE_HDR_PAGE, buf, PAGE_SIZE);
    file_hdr_ = new IxFileHdr();
    file_hdr_->deserialize(buf);

    delete[] buf;
    // disk_manager管理的fd对应的文件中，设置从file_hdr_->num_pages开始分配page_no
    int now_page_no = disk_manager_->get_fd2pageno(fd);
    disk_manager_->set_fd2pageno(fd, now_page_no + 1);
}

/**
 * @brief 用于查找指定键所在的叶子结点
 * @param key 要查找的目标key值
 * @param operation 查找到目标键值对后要进行的操作类型
 * @param transaction 事务参数，如果不需要则默认传入nullptr
 * @return [leaf node] and [root_is_latched] 返回目标叶子结点以及根结点是否加锁
 * @note need to Unlatch and unpin the leaf node outside!
 * 注意：用了FindLeafPage之后一定要unlatch叶结点，否则下次latch该结点会堵塞！
 */
std::pair<IxNodeHandle *, bool> IxIndexHandle::find_leaf_page(const char *key, Operation operation,
                                                              Transaction *transaction, bool find_first) {
    // Todo:
    // 1. 获取根节点
    // 2. 从根节点开始不断向下查找目标key
    // 3. 找到包含该key值的叶子结点停止查找，并返回叶子节点

    // page_id_t root_page_no= file_hdr_->root_page_;
    // bool root_is_latched = false;
    // IxNodeHandle *node = fetch_node(file_hdr_->root_page_);

    // if (operation == Operation::FIND) {
    //     root_latch_.unlock();
    //     node->page->RLatch();
    //     root_is_latched = false;
    // } else {
    //     node->page->WLatch();

    // }

    // root_latch_.lock();
    // if (operation != Operation::FIND) {
    //     transaction->append_index_latch_page_set(nullptr);
    // }
    bool root_is_latched = false;
    if (transaction != nullptr) {
        root_latch_.lock();
        root_is_latched = true;
    }
    
    IxNodeHandle *node = fetch_node(file_hdr_->root_page_);

    // 先考虑树的结构
    // 插入和删除可能改变树根，所以要先锁住树根
    if (transaction != nullptr) {
        if (operation == Operation::FIND) {
            node->page->RLatch();
            // std::cout << "read lock root successfully" << std::endl;
            root_latch_.unlock();
            root_is_latched = false;
        } else {
            node->page->WLatch();
            // std::cout << "insert lock root successfully" << std::endl;
            if (node->is_safe(operation)) {
                root_is_latched = false;
                release_latch_from_queue(transaction);
            }
        }
    }

    // 读操作直接加锁放锁
    // 插入和删除需要考虑孩子节点是否安全
    while (!node->is_leaf_page()) {
        page_id_t child_node_no = node->internal_lookup(key);
        IxNodeHandle* child_node = fetch_node(child_node_no);
        /* TODO */
        // 加锁
        switch (operation)
        {
        case Operation::FIND:
            if (transaction != nullptr) {
                child_node->page->RLatch();
                node->page->RUnlatch();
            }
            page_set_.insert(node->get_page_id());
            // buffer_pool_manager_->unpin_page(node->get_page_id(), false);
            break;
        case Operation::INSERT:
        case Operation::DELETE:
            if (transaction != nullptr) {
                child_node->page->WLatch();
                transaction->append_index_latch_page_set(node->page);
                if (child_node->is_safe(operation)) {
                    release_latch_from_queue(transaction);
                }
            } else {
                buffer_pool_manager_->unpin_page(node->get_page_id(), false);
            }
            break;
        default:
            break;
        }
        // buffer_pool_manager_->unpin_page(node->get_page_id(), false);
        delete node;
        node = child_node;
    }

    if (operation == Operation::FIND) {
        buffer_pool_manager_->unpin_pageSet(page_set_, false);
        page_set_.clear();
    }
    return std::make_pair(node, root_is_latched);
}

/**
 * @brief 用于查找指定键在叶子结点中的对应的值result
 *
 * @param key 查找的目标key值
 * @param result 用于存放结果的容器
 * @param transaction 事务指针
 * @return bool 返回目标键值对是否存在
 */
bool IxIndexHandle::get_value(const char *key, std::vector<Rid> *result, Transaction *transaction) {
    // Todo:
    // 1. 获取目标key值所在的叶子结点
    // 2. 在叶子节点中查找目标key值的位置，并读取key对应的rid
    // 3. 把rid存入result参数中
    // 提示：使用完buffer_pool提供的page之后，记得unpin page；记得处理并发的上锁

    auto leaf_page_pair = find_leaf_page(key, Operation::FIND, transaction);
    auto leaf_handle = leaf_page_pair.first;
    auto leaf_page_id = leaf_handle->page->get_page_id();

    Rid *value;
    bool found = leaf_handle->leaf_lookup(key, &value);

    if (found) {
        result->push_back(*value);
    }

    /* TODO */
    // 给最后一页放锁
    if (transaction != nullptr) {
        leaf_page_pair.first->page->RUnlatch();
    }
    // std::cout << "read unlock root successfully" << std::endl;
    buffer_pool_manager_->unpin_page(leaf_page_id, false);
    delete leaf_handle;
    return found;
}

/**
 * @brief  将传入的一个node拆分(Split)成两个结点，在node的右边生成一个新结点new node
 * @param node 需要拆分的结点
 * @return 拆分得到的new_node
 * @note need to unpin the new node outside
 * 注意：本函数执行完毕后，原node和new node都需要在函数外面进行unpin
 */
IxNodeHandle *IxIndexHandle::split(IxNodeHandle *node, Transaction *transaction) {
    // Todo:
    // 1. 将原结点的键值对平均分配，右半部分分裂为新的右兄弟结点
    //    需要初始化新节点的page_hdr内容
    // 2. 如果新的右兄弟结点是叶子结点，更新新旧节点的prev_leaf和next_leaf指针
    //    为新节点分配键值对，更新旧节点的键值对数记录
    // 3. 如果新的右兄弟结点不是叶子结点，更新该结点的所有孩子结点的父节点信息(使用IxIndexHandle::maintain_child())

    // 每个节点至少保留的节点数量
    int min_size = std::ceil(node->get_size() / 2);
    // 对新节点进行初始化
    auto new_node_handle = create_node();
    *new_node_handle->page_hdr = {
        .next_free_page_no = node->page_hdr->next_free_page_no,
        .parent = node->get_parent_page_no(),
        .num_key = 0,
        .is_leaf = node->page_hdr->is_leaf
    };
//    new_node_handle->page_hdr->next_free_page_no = node->page_hdr->next_free_page_no;
//    new_node_handle->page_hdr->parent = node->get_parent_page_no();
//    new_node_handle->page_hdr->num_key = 0;
//    new_node_handle->page_hdr->is_leaf = node->page_hdr->is_leaf;
    if (node->is_leaf_page()) {
        // 链表插入
        // 新节点的下一个节点是前节点的下一个节点
        new_node_handle->page_hdr->next_leaf = node->page_hdr->next_leaf;

        // 下一个页的前节点就是自己
        auto next_leaf = fetch_node(node->page_hdr->next_leaf);
        // TODO 加锁
        if (transaction != nullptr) {
            next_leaf->page->WLatch();
        }
        
        next_leaf->page_hdr->prev_leaf = new_node_handle->get_page_no();

        if (transaction != nullptr) {
            next_leaf->page->WUnlatch();
        }
        
        buffer_pool_manager_->unpin_page(next_leaf->get_page_id(), true);
        delete next_leaf;
        // 前一个节点的后一个节点是自己
        node->page_hdr->next_leaf = new_node_handle->get_page_no();
        // 自己的前一个节点
        new_node_handle->page_hdr->prev_leaf = node->get_page_no();
    }


    new_node_handle->insert_pairs(0, node->get_key(min_size), node->get_rid(min_size), node->get_size() - min_size);
    // 注意，internal节点的第一个键是无意义的

    node->set_size(min_size);
    if (!node->is_leaf_page()) {
        for (int i = 0; i < new_node_handle->page_hdr->num_key; i++){
            maintain_child(new_node_handle, i);
        }
    }

    return new_node_handle;
}

/**
 * @brief Insert key & value pair into internal page after split
 * 拆分(Split)后，向上找到old_node的父结点
 * 将new_node的第一个key插入到父结点，其位置在 父结点指向old_node的孩子指针 之后
 * 如果插入后>=maxsize，则必须继续拆分父结点，然后在其父结点的父结点再插入，即需要递归
 * 直到找到的old_node为根结点时，结束递归（此时将会新建一个根R，关键字为key，old_node和new_node为其孩子）
 *
 * @param (old_node, new_node) 原结点为old_node，old_node被分裂之后产生了新的右兄弟结点new_node
 * @param key 要插入parent的key
 * @note 一个结点插入了键值对之后需要分裂，分裂后左半部分的键值对保留在原结点，在参数中称为old_node，
 * 右半部分的键值对分裂为新的右兄弟节点，在参数中称为new_node（参考Split函数来理解old_node和new_node）
 * @note 本函数执行完毕后，new node和old node都需要在函数外面进行unpin
 */
void IxIndexHandle::insert_into_parent(IxNodeHandle *old_node, const char *key, IxNodeHandle *new_node,
                                       Transaction *transaction) {
    // Todo:
    // 1. 分裂前的结点（原结点, old_node）是否为根结点，如果为根结点需要分配新的root
    // 2. 获取原结点（old_node）的父亲结点
    // 3. 获取key对应的rid，并将(key, rid)插入到父亲结点
    // 4. 如果父亲结点仍需要继续分裂，则进行递归插入
    // 提示：记得unpin page

    if (old_node->is_root_page()) {
        // 初始化根节点
        IxNodeHandle *new_root_page = create_node();
        file_hdr_->root_page_ = new_root_page->get_page_no();
        *new_root_page->page_hdr = {
                .next_free_page_no = IX_NO_PAGE,
                .parent = IX_NO_PAGE,
                .num_key = 0,
                .is_leaf = false,
                .prev_leaf = IX_LEAF_HEADER_PAGE,
                .next_leaf = IX_LEAF_HEADER_PAGE,
        };

        old_node->set_parent_page_no(new_root_page->get_page_no());
        new_node->set_parent_page_no(new_root_page->get_page_no());

        new_root_page->insert_pair(0, old_node->get_key(0), {old_node->get_page_no(), 0});
        new_root_page->insert_pair(1, key, {new_node->get_page_no(), 0});

//        new_root_page->set_key(0, old_node->get_key(0)); // 下标为0的键是无效的
//        new_root_page->set_rid(0, Rid{old_node->get_page_no(), 0});
//        new_root_page->set_key(1, key);
//        new_root_page->set_rid(1, Rid{new_node->get_page_no(), 0});
//        new_root_page->set_size(2);

        buffer_pool_manager_->unpin_page(new_root_page->get_page_id(), true);
        delete new_root_page;
        return;
    }

    // 找到父节点
    auto parent_node = fetch_node(old_node->get_parent_page_no());
    new_node->set_parent_page_no(parent_node->get_page_no());

    int index = parent_node->upper_bound(key);
    parent_node->insert_pair(index, key, {new_node->get_page_no(), 0});
    if(parent_node->get_size() >= parent_node->get_max_size()){
        // 继续分裂
        IxNodeHandle* uncle_node = split(parent_node, transaction);
        insert_into_parent(parent_node, uncle_node->get_key(0), uncle_node, transaction);
        buffer_pool_manager_->unpin_page(uncle_node->get_page_id(), true);
        delete uncle_node;
    }

    if(transaction != nullptr){
        release_latch_from_queue(transaction);
    }
    //release_latch_from_queue(transaction);
    buffer_pool_manager_->unpin_page(parent_node->get_page_id(), true);
    delete parent_node;
}

/**
 * @brief 将指定键值对插入到B+树中
 * @param (key, value) 要插入的键值对
 * @param transaction 事务指针
 * @return page_id_t 插入到的叶结点的page_no
 */
page_id_t IxIndexHandle::insert_entry(const char *key, const Rid &value, Transaction *transaction) {
    // Todo:
    // 1. 查找key值应该插入到哪个叶子节点
    // 2. 在该叶子节点中插入键值对
    // 3. 如果结点已满，分裂结点，并把新结点的相关信息插入父节点
    // 提示：记得unpin page；若当前叶子节点是最右叶子节点，则需要更新file_hdr_.last_leaf；记得处理并发的上锁
    if (transaction != nullptr) {
        transaction->append_index_latch_page_set(nullptr);
    }
    auto [leaf_page_handle, latch] = find_leaf_page(key, Operation::INSERT, transaction);

    if (leaf_page_handle == nullptr) {
        leaf_page_handle = create_node();
        file_hdr_->root_page_ = leaf_page_handle->get_page_no();
        file_hdr_->first_leaf_ = leaf_page_handle->get_page_no();
        file_hdr_->last_leaf_ = leaf_page_handle->get_page_no();
        leaf_page_handle->page_hdr->next_free_page_no = IX_NO_PAGE;
        leaf_page_handle->page_hdr->prev_leaf = IX_LEAF_HEADER_PAGE;
    }

    int prev_size = leaf_page_handle->get_size();
    int size = leaf_page_handle->insert(key, value);

    if (size == prev_size) {
        // 没插入
        if (transaction != nullptr) {
            release_latch_from_queue(transaction);
            leaf_page_handle->page->WUnlatch();
        }
        buffer_pool_manager_->unpin_page(leaf_page_handle->get_page_id(), true);
        delete leaf_page_handle;
        return -1;
    }

    int index = leaf_page_handle->lower_bound(key);
    if (size >= leaf_page_handle->get_max_size()) {
        IxNodeHandle *new_leaf_page = split(leaf_page_handle, transaction);
        insert_into_parent(leaf_page_handle, new_leaf_page->get_key(0), new_leaf_page, transaction);
        // 记得把split出来的节点unpin
        buffer_pool_manager_->unpin_page(new_leaf_page->get_page_id(), true);
        // 更新最后的叶节点和最前的叶节点
        if (leaf_page_handle->get_page_no() == file_hdr_->last_leaf_) {
            file_hdr_->last_leaf_ = new_leaf_page->get_page_no();
        }
        delete new_leaf_page;
    }

    if (transaction != nullptr) {
        release_latch_from_queue(transaction);
        leaf_page_handle->page->WUnlatch();
    }
    auto page_id = leaf_page_handle->get_page_id();
    buffer_pool_manager_->unpin_page(leaf_page_handle->get_page_id(), true);
    delete leaf_page_handle;
    return page_id.page_no;
}

/**
 * @brief 用于删除B+树中含有指定key的键值对
 * @param key 要删除的key值
 * @param transaction 事务指针
 */
bool IxIndexHandle::delete_entry(const char *key, Transaction *transaction) {
    // Todo:
    // 1. 获取该键值对所在的叶子结点
    // 2. 在该叶子结点中删除键值对
    // 3. 如果删除成功需要调用CoalesceOrRedistribute来进行合并或重分配操作，并根据函数返回结果判断是否有结点需要删除
    // 4. 如果需要并发，并且需要删除叶子结点，则需要在事务的delete_page_set中添加删除结点的对应页面；记得处理并发的上锁
    transaction->append_index_latch_page_set(nullptr);
    auto [leaf_page_handle, latch] = find_leaf_page(key, Operation::DELETE, transaction);
    int current_size = leaf_page_handle->get_size();
    int size = leaf_page_handle->remove(key);

    bool delete_node = false;
    if (size < current_size) {
        // 删除成功
        delete_node = coalesce_or_redistribute(leaf_page_handle, transaction, &latch);
        release_latch_from_queue(transaction);
        if (delete_node) {
            // 该节点会被删除
            transaction->append_index_deleted_page(leaf_page_handle->page);
        }
        for (auto &page : *(transaction->get_index_deleted_page_set())) {
            buffer_pool_manager_->delete_page(page->get_page_id());
        }

        transaction->get_index_deleted_page_set()->clear();
        leaf_page_handle->page->WUnlatch();
        buffer_pool_manager_->unpin_page(leaf_page_handle->get_page_id(), true);
        delete leaf_page_handle;
        return true;
    }

    // 删除失败
    release_latch_from_queue(transaction);
    leaf_page_handle->page->WUnlatch();
    buffer_pool_manager_->unpin_page(leaf_page_handle->get_page_id(), true);
    delete leaf_page_handle;
    return false;
}

/**
 * @brief 用于处理合并和重分配的逻辑，用于删除键值对后调用
 *
 * @param node 执行完删除操作的结点
 * @param transaction 事务指针
 * @param root_is_latched 传出参数：根节点是否上锁，用于并发操作
 * @return 是否需要删除结点
 * @note User needs to first find the sibling of input page.
 * If sibling's size + input page's size >= 2 * page's minsize, then redistribute.
 * Otherwise, merge(Coalesce).
 */
bool IxIndexHandle::coalesce_or_redistribute(IxNodeHandle *node, Transaction *transaction, bool *root_is_latched) {
    // Todo:
    // 1. 判断node结点是否为根节点
    //    1.1 如果是根节点，需要调用AdjustRoot() 函数来进行处理，返回根节点是否需要被删除
    //    1.2 如果不是根节点，并且不需要执行合并或重分配操作，则直接返回false，否则执行2
    // 2. 获取node结点的父亲结点
    // 3. 寻找node结点的兄弟结点（优先选取前驱结点）
    // 4. 如果node结点和兄弟结点的键值对数量之和，能够支撑两个B+树结点（即node.size+neighbor.size >=
    // NodeMinSize*2)，则只需要重新分配键值对（调用Redistribute函数）
    // 5. 如果不满足上述条件，则需要合并两个结点，将右边的结点合并到左边的结点（调用Coalesce函数）

    if (file_hdr_->root_page_ == node->get_page_no()) {
        // 如果是根节点，调整根
        if (adjust_root(node)) {
            transaction->append_index_deleted_page(node->page);
            return true;
        }
        return false;
    }

    // 节点满足要求
    if (node->get_size() >= node->get_min_size()) {
        //maintain_parent(node);
        return false;
    }

    auto parent_handle = fetch_node(node->get_parent_page_no()); // 获取要处理的节点的父节点

    // 找到子节点在哪
    int index = -1;
    for (int i = 0; i < parent_handle->get_size(); i++) {
        if (parent_handle->get_rid(i)->page_no == node->get_page_no()) {
            index = i;
            break;
        }
    }

    IxNodeHandle *sibling; // 尽量找前驱节点
    if (index > 0) {
        // 找前驱节点
        sibling = fetch_node(parent_handle->get_rid(index - 1)->page_no);
    } else {
        sibling = fetch_node(parent_handle->get_rid(index + 1)->page_no);
    }

    if (node->get_size() + sibling->get_size() >= node->get_max_size() * 2) {
        // 一个节点放不下，则合并
        // 这种情况没有任何节点会被删除
        redistribute(sibling, node, parent_handle, index);
        buffer_pool_manager_->unpin_page(sibling->get_page_id(), true);
        buffer_pool_manager_->unpin_page(parent_handle->get_page_id(), true);
        delete sibling;
        delete parent_handle;
        return false;
    }

    bool parent_should_delete = coalesce(&sibling, &node, &parent_handle, index, transaction, root_is_latched);
    if (parent_should_delete) {
        transaction->append_index_deleted_page(parent_handle->page);
    }

    if (index == 0) {
        // 此时邻居节点会被删除
        transaction->append_index_deleted_page(sibling->page);
    }
    buffer_pool_manager_->unpin_page(sibling->get_page_id(), true);
    delete sibling;
    buffer_pool_manager_->unpin_page(parent_handle->get_page_id(), true);
    delete parent_handle;
    return index > 0;
}

/**
 * @brief 用于当根结点被删除了一个键值对之后的处理
 * @param old_root_node 原根节点
 * @return bool 根结点是否需要被删除
 * @note size of root page can be less than min size and this method is only called within coalesce_or_redistribute()
 */
bool IxIndexHandle::adjust_root(IxNodeHandle *old_root_node) {
    // Todo:
    // 1. 如果old_root_node是内部结点，并且大小为1，则直接把它的孩子更新成新的根结点
    // 2. 如果old_root_node是叶结点，且大小为0，则直接更新root page
    // 3. 除了上述两种情况，不需要进行操作

    if (!old_root_node->is_leaf_page() && old_root_node->get_size() == 1) {
        auto child_page_id = old_root_node->remove_and_return_only_child();
        file_hdr_->root_page_ = child_page_id;
        auto child_page = fetch_node(child_page_id);
        child_page->set_parent_page_no(INVALID_PAGE_ID);
        buffer_pool_manager_->unpin_page(child_page->get_page_id(), true);
        release_node_handle(*old_root_node);
        delete child_page;
        return true;
    }

    if (old_root_node->is_leaf_page() && old_root_node->get_size() == 0) {
        file_hdr_->root_page_ = INVALID_PAGE_ID;
        disk_manager_->set_fd2pageno(fd_, IX_INIT_ROOT_PAGE);
        return true;
    }
    return false;
}

/**
 * @brief 重新分配node和兄弟结点neighbor_node的键值对
 * Redistribute key & value pairs from one page to its sibling page. If index == 0, move sibling page's first key
 * & value pair into end of input "node", otherwise move sibling page's last key & value pair into head of input "node".
 *
 * @param neighbor_node sibling page of input "node"
 * @param node input from method coalesceOrRedistribute()
 * @param parent the parent of "node" and "neighbor_node"
 * @param index node在parent中的rid_idx
 * @note node是之前刚被删除过一个key的结点
 * index=0，则neighbor是node后继结点，表示：node(left)      neighbor(right)
 * index>0，则neighbor是node前驱结点，表示：neighbor(left)  node(right)
 * 注意更新parent结点的相关kv对
 */
void IxIndexHandle::redistribute(IxNodeHandle *neighbor_node, IxNodeHandle *node, IxNodeHandle *parent, int index) {
    // Todo:
    // 1. 通过index判断neighbor_node是否为node的前驱结点
    // 2. 从neighbor_node中移动一个键值对到node结点中
    // 3. 更新父节点中的相关信息，并且修改移动键值对对应孩字结点的父结点信息（maintain_child函数）
    // 注意：neighbor_node的位置不同，需要移动的键值对不同，需要分类讨论

    if (index != 0) {
        // 找的是前节点
        if (node->is_leaf_page()) {
            char *key = neighbor_node->get_key(neighbor_node->get_size() - 1);
            Rid *value = neighbor_node->get_rid(neighbor_node->get_size() - 1);
            node->insert_pair(0, key, *value);
            neighbor_node->erase_pair(neighbor_node->get_size() - 1);
        } else {
            // 因为中间节点第一个键是无效的，所以要将父节点对应位置的键挪下来
            node->set_key(0, parent->get_key(index));
            node->insert_pair(0, neighbor_node->get_key(neighbor_node->get_size() - 1), *neighbor_node->get_rid(neighbor_node->get_size() - 1));
            neighbor_node->erase_pair(neighbor_node->get_size() - 1);
            maintain_child(node, 0);
        }
        parent->set_key(index, node->get_key(0));
        return;
    }

    if (node->is_leaf_page()) {
        char *key = neighbor_node->get_key(0);
        Rid *value = neighbor_node->get_rid(0);
        node->insert_pair(node->get_size(), key, *value);
        neighbor_node->erase_pair(0);
    } else {
        // 为了键的正确性，需要将父节点对应位置的键挪下来，并用后一个节点的键替补上去
        node->insert_pair(node->get_size(), parent->get_key(1), *neighbor_node->get_rid(0));
        neighbor_node->erase_pair(0);
        maintain_child(node, node->get_size() - 1);
    }
    parent->set_key(1, neighbor_node->get_key(0));
}

/**
 * @brief 合并(Coalesce)函数是将node和其直接前驱进行合并，也就是和它左边的neighbor_node进行合并；
 * 假设node一定在右边。如果上层传入的index=0，说明node在左边，那么交换node和neighbor_node，保证node在右边；合并到左结点，实际上就是删除了右结点；
 * Move all the key & value pairs from one page to its sibling page, and notify buffer pool manager to delete this page.
 * Parent page must be adjusted to take info of deletion into account. Remember to deal with coalesce or redistribute
 * recursively if necessary.
 *
 * @param neighbor_node sibling page of input "node" (neighbor_node是node的前结点)
 * @param node input from method coalesceOrRedistribute() (node结点是需要被删除的)
 * @param parent parent page of input "node"
 * @param index node在parent中的rid_idx
 * @return true means parent node should be deleted, false means no deletion happend
 * @note Assume that *neighbor_node is the left sibling of *node (neighbor -> node)
 */
bool IxIndexHandle::coalesce(IxNodeHandle **neighbor_node, IxNodeHandle **node, IxNodeHandle **parent, int index,
                             Transaction *transaction, bool *root_is_latched) {
    // Todo:
    // 1. 用index判断neighbor_node是否为node的前驱结点，若不是则交换两个结点，让neighbor_node作为左结点，node作为右结点
    // 2. 把node结点的键值对移动到neighbor_node中，并更新node结点孩子结点的父节点信息（调用maintain_child函数）
    // 3. 释放和删除node结点，并删除parent中node结点的信息，返回parent是否需要被删除
    // 提示：如果是叶子结点且为最右叶子结点，需要更新file_hdr_.last_leaf

    IxNodeHandle *left, *right;
    if (index == 0) {
        // 交换顺序，保持统一
        left = *node;
        right = *neighbor_node;
        index = 1;
    } else {
        left = *neighbor_node;
        right = *node;
    }

    auto key_in_parent = (*parent)->get_key(index);
    if (right->is_leaf_page()) {
        int node_size = right->get_size();
        for (int i = 0; i < node_size; i++) {
            auto key = right->get_key(i);
            auto value = right->get_rid(i);
            left->insert_pair(left->get_size(), key, *value);
        }
        if (file_hdr_->last_leaf_ == right->get_page_no()) {
            file_hdr_->last_leaf_ = left->get_page_no();
        }
        erase_leaf(right);
    } else {
        right->set_key(0, key_in_parent);
        left->insert_pairs(left->get_size(), right->keys, (right->rids), right->get_size());
        for (int i = 0; i < left->get_size(); i++){
            maintain_child(left, i);
        }
    }

    (*parent)->erase_pair(index);
    release_node_handle(*right);
    delete left;
    delete right;
    return coalesce_or_redistribute(*parent, transaction, root_is_latched);
}

/**
 * @brief 这里把iid转换成了rid，即iid的slot_no作为node的rid_idx(key_idx)
 * node其实就是把slot_no作为键值对数组的下标
 * 换而言之，每个iid对应的索引槽存了一对(key,rid)，指向了(要建立索引的属性首地址,插入/删除记录的位置)
 *
 * @param iid
 * @return Rid
 * @note iid和rid存的不是一个东西，rid是上层传过来的记录位置，iid是索引内部生成的索引槽位置
 */
Rid IxIndexHandle::get_rid(const Iid &iid) const {
    IxNodeHandle *node = fetch_node(iid.page_no);
    if (iid.slot_no >= node->get_size()) {
        throw IndexEntryNotFoundError();
    }
    buffer_pool_manager_->unpin_page(node->get_page_id(), false);  // unpin it!
    auto res_rid = *node->get_rid(iid.slot_no);
    delete node;
    return res_rid;
}

/**
 * @brief FindLeafPage + lower_bound
 *
 * @param key
 * @return Iid
 * @note 上层传入的key本来是int类型，通过(const char *)&key进行了转换
 * 可用*(int *)key转换回去
 */
Iid IxIndexHandle::lower_bound(const char *key) {
    auto [leaf_page, root_is_latched] = find_leaf_page(key, Operation::FIND, nullptr);
    Iid ret;
    ret.page_no = leaf_page->get_page_no();
    ret.slot_no = leaf_page->lower_bound(key);

    if (ret.slot_no == leaf_page->get_size() && ret.page_no != file_hdr_->last_leaf_ ) {
        ret.page_no = leaf_page->get_next_leaf();
        ret.slot_no = 0;
    }
    leaf_page->page->RUnlatch();
    buffer_pool_manager_->unpin_page(leaf_page->get_page_id(), false);
    delete leaf_page;
    return ret;
}

/**
 * @brief FindLeafPage + upper_bound
 *
 * @param key
 * @return Iid
 * @brief 特殊情况，该页面没有值，即页面为空，这时upper_bound的返回值slot将大于num_key
 */
Iid IxIndexHandle::upper_bound(const char *key) {
    auto [leaf_page, root_is_latched] = find_leaf_page(key, Operation::FIND, nullptr);
    Iid ret;
    ret.page_no = leaf_page->get_page_no();
    ret.slot_no = leaf_page->upper_bound(key);
    if (ret.page_no != file_hdr_->last_leaf_ && ret.slot_no == leaf_page->get_size()) {
        ret.page_no = leaf_page->get_next_leaf();
        ret.slot_no = 0;
    }
    leaf_page->page->RUnlatch();
    buffer_pool_manager_->unpin_page(leaf_page->get_page_id(), false);
    delete leaf_page;
    return ret;
}

/**
 * @brief 指向最后一个叶子的最后一个结点的后一个
 * 用处在于可以作为IxScan的最后一个
 *
 * @return Iid
 */
Iid IxIndexHandle::leaf_end() const {
    IxNodeHandle *node = fetch_node(file_hdr_->last_leaf_);
    Iid iid = {.page_no = file_hdr_->last_leaf_, .slot_no = node->get_size()};
    buffer_pool_manager_->unpin_page(node->get_page_id(), false);  // unpin it!
    delete node;
    return iid;
}

/**
 * @brief 指向第一个叶子的第一个结点
 * 用处在于可以作为IxScan的第一个
 *
 * @return Iid
 */
Iid IxIndexHandle::leaf_begin() const {
    Iid iid = {.page_no = file_hdr_->first_leaf_, .slot_no = 0};
    return iid;
}

/**
 * @brief 获取一个指定结点
 *
 * @param page_no
 * @return IxNodeHandle*
 * @note pin the page, remember to unpin it outside!
 */
IxNodeHandle *IxIndexHandle::fetch_node(int page_no) const {
    assert(page_no != INVALID_PAGE_ID);
    Page *page = buffer_pool_manager_->fetch_page(PageId{fd_, page_no});
    IxNodeHandle *node = new IxNodeHandle(file_hdr_, page);

    return node;
}

/**
 * @brief 创建一个新结点
 *
 * @return IxNodeHandle*
 * @note pin the page, remember to unpin it outside!
 * 注意：对于Index的处理是，删除某个页面后，认为该被删除的页面是free_page
 * 而first_free_page实际上就是最新被删除的页面，初始为IX_NO_PAGE
 * 在最开始插入时，一直是create node，那么first_page_no一直没变，一直是IX_NO_PAGE
 * 与Record的处理不同，Record将未插入满的记录页认为是free_page
 */
IxNodeHandle *IxIndexHandle::create_node() {
    IxNodeHandle *node;
    file_hdr_->num_pages_++;
    PageId new_page_id(fd_,INVALID_PAGE_ID);
    //PageId new_page_id = {.fd = fd_, .page_no = INVALID_PAGE_ID};
    // 从3开始分配page_no，第一次分配之后，new_page_id.page_no=3，file_hdr_.num_pages=4
    Page *page = buffer_pool_manager_->new_page(&new_page_id);
    node = new IxNodeHandle(file_hdr_, page);
    return node;
}

/**
 * @brief 从node开始更新其父节点的第一个key，一直向上更新直到根节点
 *
 * @param node
 */
void IxIndexHandle::maintain_parent(IxNodeHandle *node) {
    IxNodeHandle *curr = node;
    while (curr->get_parent_page_no() != IX_NO_PAGE) {
        // Load its parent
        IxNodeHandle *parent = fetch_node(curr->get_parent_page_no());
        int rank = parent->find_child(curr);
        char *parent_key = parent->get_key(rank);
        char *child_first_key = curr->get_key(0);
        if (memcmp(parent_key, child_first_key, file_hdr_->col_tot_len_) == 0) {
            assert(buffer_pool_manager_->unpin_page(parent->get_page_id(), true));
            break;
        }
        memcpy(parent_key, child_first_key, file_hdr_->col_tot_len_);  // 修改了parent node
        curr = parent;

        assert(buffer_pool_manager_->unpin_page(parent->get_page_id(), true));
    }
}

/**
 * @brief 要删除leaf之前调用此函数，更新leaf前驱结点的next指针和后继结点的prev指针
 *
 * @param leaf 要删除的leaf
 */
void IxIndexHandle::erase_leaf(IxNodeHandle *leaf) {
    assert(leaf->is_leaf_page());

    IxNodeHandle *prev = fetch_node(leaf->get_prev_leaf());
    prev->set_next_leaf(leaf->get_next_leaf());
    buffer_pool_manager_->unpin_page(prev->get_page_id(), true);
    delete prev;

    IxNodeHandle *next = fetch_node(leaf->get_next_leaf());
    next->set_prev_leaf(leaf->get_prev_leaf());  // 注意此处是SetPrevLeaf()
    buffer_pool_manager_->unpin_page(next->get_page_id(), true);
    delete next;
}

/**
 * @brief 删除node时，更新file_hdr_.num_pages
 *
 * @param node
 */
void IxIndexHandle::release_node_handle(IxNodeHandle &node) {
    file_hdr_->num_pages_--;
}

/**
 * @brief 将node的第child_idx个孩子结点的父节点置为node
 */
void IxIndexHandle::maintain_child(IxNodeHandle *node, int child_idx) {
    if (!node->is_leaf_page()) {
        //  Current node is inner node, load its child and set its parent to current node
        int child_page_no = node->value_at(child_idx);
        IxNodeHandle *child = fetch_node(child_page_no);
        child->set_parent_page_no(node->get_page_no());
        buffer_pool_manager_->unpin_page(child->get_page_id(), true);
        delete child;
    }
}

page_id_t IxIndexHandle::load_insert_entry(const char *key, const Rid &value,std::vector<PageId>& unpin_index_page) {
    //std::set<PageId> unpin_page_set;
    IxNodeHandle* node = fetch_node(file_hdr_->root_page_);
    while (!node->is_leaf_page()){
        page_id_t child_node_no = node->internal_lookup(key);
        IxNodeHandle* child_node = fetch_node(child_node_no);
        //TODO unpin
//        buffer_pool_manager_->unpin_page(interval_handle->get_page_id(), false);
        //unpin_page_set.insert(node->get_page_id());
        unpin_index_page.emplace_back(node->get_page_id());
        delete node;
        node = child_node;
    }
    buffer_pool_manager_->unpin_pageSet(unpin_index_page,false);
    unpin_index_page.clear();
    //buffer_pool_manager_->unpin_pageSet(unpin_page_set,false);//unpin非叶子节点

    page_id_t page_no = node->get_page_no();
    int prev_size = node->get_size();
    int size = node->insert(key, value);
    if (size == prev_size) {
        // 没插入
        buffer_pool_manager_->unpin_page(node->get_page_id(), false);
        delete node;
        return -1;
    }

    if (size >= node->get_max_size()) {
        auto new_right_node_handle = single_split(node);
        single_insert_into_parent(node, new_right_node_handle->get_key(0), new_right_node_handle);
        if (node->get_page_no() == file_hdr_->last_leaf_) {
            file_hdr_->last_leaf_ = new_right_node_handle->get_page_no();
        }
        buffer_pool_manager_->unpin_page(new_right_node_handle->get_page_id(),true);
        delete new_right_node_handle;
    }
    buffer_pool_manager_->unpin_page(node->get_page_id(),true);
    delete node;
    return page_no;
}

IxNodeHandle *IxIndexHandle::single_split(IxNodeHandle *node) {
    int min_size = std::ceil(node->get_size() / 2);

    auto new_node_handle = create_node();
    new_node_handle->page_hdr->next_free_page_no = node->page_hdr->next_free_page_no;
    new_node_handle->page_hdr->parent = node->get_parent_page_no();
    new_node_handle->page_hdr->num_key = 0;

    if (node->is_leaf_page()) {
        new_node_handle->page_hdr->is_leaf = true;
        // 链表插入
        // 新节点的下一个节点是前节点的下一个节点
        new_node_handle->page_hdr->next_leaf = node->page_hdr->next_leaf;

        // 下一个页的前节点就是自己
        auto next_leaf = fetch_node(node->page_hdr->next_leaf);
        next_leaf->page_hdr->prev_leaf = new_node_handle->get_page_no();
        buffer_pool_manager_->unpin_page(next_leaf->get_page_id(), true);
        delete next_leaf;
        // 前一个节点的后一个节点是自己
        node->page_hdr->next_leaf = new_node_handle->get_page_no();
        // 自己的前一个节点
        new_node_handle->page_hdr->prev_leaf = node->get_page_no();        
    } else {
        new_node_handle->page_hdr->is_leaf = false;
    }
    
    new_node_handle->insert_pairs(0, node->get_key(min_size), node->get_rid(min_size), node->get_size() - min_size);
    // 注意，internal节点的第一个键是无意义的

    node->set_size(min_size);
    if (!node->is_leaf_page()) {
        for (int i = 0; i < new_node_handle->page_hdr->num_key; i++){
            maintain_child(new_node_handle, i);
        }
    }

    return new_node_handle;
}

void IxIndexHandle::single_insert_into_parent(IxNodeHandle *old_node, const char *key, IxNodeHandle *new_node) {
    if (old_node->is_root_page()) {
        // 初始化根节点
        IxNodeHandle *new_root_page = create_node();
        file_hdr_->root_page_ = new_root_page->get_page_no();
        new_root_page->page_hdr->num_key = 0;
        new_root_page->page_hdr->is_leaf = false;

        new_root_page->page_hdr->next_free_page_no = IX_NO_PAGE;
        new_root_page->page_hdr->parent = IX_NO_PAGE;
        new_root_page->page_hdr->prev_leaf = IX_LEAF_HEADER_PAGE;
        new_root_page->page_hdr->next_leaf = IX_LEAF_HEADER_PAGE;

        old_node->set_parent_page_no(new_root_page->get_page_no());
        new_node->set_parent_page_no(new_root_page->get_page_no());

        new_root_page->set_key(0, old_node->get_key(0)); // 下标为0的键是无效的
        new_root_page->set_rid(0, Rid{old_node->get_page_no(), 0});
        new_root_page->set_key(1, key);
        new_root_page->set_rid(1, Rid{new_node->get_page_no(), 0});
        new_root_page->set_size(2);

        buffer_pool_manager_->unpin_page(new_root_page->get_page_id(), true);
        delete new_root_page;
        return;
    }

    // 找到父节点
    auto parent_node = fetch_node(old_node->get_parent_page_no());
    new_node->set_parent_page_no(parent_node->get_page_no());

    int index = parent_node->upper_bound(key);
    parent_node->insert_pair(index, key, {new_node->get_page_no(), 0});
    if (parent_node->get_size() >= parent_node->get_max_size()) {
        // 继续分裂
        IxNodeHandle* uncle_node = single_split(parent_node);
        single_insert_into_parent(parent_node, uncle_node->get_key(0), uncle_node);
        buffer_pool_manager_->unpin_page(uncle_node->get_page_id(), true);
        delete uncle_node;
    }

    buffer_pool_manager_->unpin_page(parent_node->get_page_id(), true);
    delete parent_node;
}


/**
 * @brief 释放队列中的所有锁
*/
void IxIndexHandle::release_latch_from_queue(Transaction *transaction) {
    while (!transaction->get_index_latch_page_set()->empty()) {
        Page *page = transaction->get_index_latch_page_set()->front();
        transaction->get_index_latch_page_set()->pop_front();
        if (page == nullptr) {
            this->root_latch_.unlock();
        } else {
            page->WUnlatch();
            buffer_pool_manager_->unpin_page(page->get_page_id(), false);
        }
    }
}

int IxIndexHandle::first_leaf_first_key() const {
    IxNodeHandle *node = fetch_node(file_hdr_->first_leaf_);
    assert(file_hdr_->col_types_[0] == ColType::TYPE_INT);
    if(node->get_size() == 0){
        return 1;
    }
    int res = *(int*)(node->get_key(0));
    buffer_pool_manager_->unpin_page(node->get_page_id(),false);
    delete node;
    return res;
}


int IxIndexHandle::last_leaf_last_key() const {
    IxNodeHandle *node = fetch_node(file_hdr_->last_leaf_);
    assert(file_hdr_->col_types_[0] == ColType::TYPE_INT);
    if(node->get_size() == 0){
        return 0;
    }
    int res = *(int*)(node->get_key(node->get_size() - 1));
    buffer_pool_manager_->unpin_page(node->get_page_id(),false);
    delete node;
    return res;
}

bool IxIndexHandle::find_ind_key_at(const Iid& iid_, int& ret) const{
    if(iid_.page_no == -1){
        return false;
    }
    IxNodeHandle *node = fetch_node(iid_.page_no);
    if(iid_.slot_no >= node->get_size()){
        return false;
    }
    ret = *(int*)(node->get_key(iid_.slot_no));

    delete node;
    return true;
}


/**
 * @breif 用于间隙锁的间隙
 * @param key 插入的key
 * @param num_key key所在的页内的，插入前的key的数量
 * @return 第一个大于key的位置
 */
Iid IxIndexHandle::upper_bound(const char *key,Rid &rid) {
    auto [leaf_page_handle, latch] = find_leaf_page(key, Operation::FIND, nullptr);
//    num_key = leaf_page_handle->page_hdr->num_key;
    int index = leaf_page_handle->upper_bound(key);
    if (index == leaf_page_handle->get_size()) {
        leaf_page_handle->page->RUnlatch();
        std::cout << "read unlock root successfully" << std::endl;
        buffer_pool_manager_->unpin_page(leaf_page_handle->get_page_id(), false);
        rid = {-1,-1};
        return leaf_end();
    }
    if(ix_compare(leaf_page_handle->get_key(0),key,leaf_page_handle->file_hdr->col_types_,leaf_page_handle->file_hdr->col_lens_)==1){
        //key小
        rid = *leaf_page_handle->get_rid(0);
        index = 0;
    }else{
        rid = *leaf_page_handle->get_rid(index);
    }
    auto page_no = leaf_page_handle->get_page_no();
    leaf_page_handle->page->RUnlatch();
    std::cout << "read unlock root successfully" << std::endl;
    buffer_pool_manager_->unpin_page(leaf_page_handle->get_page_id(), false);
    return Iid{page_no, index};
}

/**
 *
 * @param left scan的起始点
 * @param right 终点,如果1,2,3,4,select <=3,那么4是终点
 * @param context_ 上下文
 * @brief 该函数不处理等值查询，因此需要程序员在本函数外部实现
 */
void IxIndexHandle::gap_lock(Iid left, Iid right,Context *context_,int table_fd) {
    std::unique_lock<std::mutex> l(root_latch_);
    //assert(left.page_no == this->file_hdr_->first_leaf_);会报错
    auto node = fetch_node(left.page_no);
    buffer_pool_manager_->unpin_page(node->get_page_id(), false);
    //case1:间隙锁在最小索引前
    //case2:gap lock 在 最大索引后
    if(right.page_no == file_hdr_->first_leaf_ && right.slot_no == 0){
        context_->lock_mgr_->lock_gap_on_index(context_->txn_,this->fd_,*node->get_rid(0));
        buffer_pool_manager_->unpin_page(node->get_page_id(), false);
        return;
    }
    if(left.page_no == file_hdr_->last_leaf_ && left.slot_no == node->get_size()){
        context_->lock_mgr_->lock_gap_on_index(context_->txn_,this->fd_,{-1,-1});
        buffer_pool_manager_->unpin_page(node->get_page_id(), false);
        return;
    }
    //范围查询
    Iid last;
    Iid cur = left;
    do{
        if(cur.page_no == file_hdr_->last_leaf_  && cur.slot_no == node->get_size()){
            // 最后一个节点, 对最后一个间隙上间隙锁, {-1， -1}标识为最后一个间隙
            context_->lock_mgr_->lock_gap_on_index(context_->txn_, this->fd_,{-1, -1});
            buffer_pool_manager_->unpin_page(node->get_page_id(), false);
            break;
        }
        else{
            context_->lock_mgr_->lock_gap_on_index(context_->txn_, this->fd_,*node->get_rid(cur.slot_no));

            //context_->lock_mgr_->lock_shared_on_record(context_->txn_, *node->get_rid(cur.slot_no), table_fd);
        }
        last = cur;
        // increment slot no
        cur.slot_no++;
        if (cur.page_no != file_hdr_->last_leaf_ && cur.slot_no == node->get_size()){
            //assert(0);
            cur.slot_no = 0;
            cur.page_no = node->get_next_leaf();
            buffer_pool_manager_->unpin_page(node->get_page_id(), false);
            node = fetch_node(cur.page_no);
        }
    } while (last != right);


}
