/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "rm_file_handle.h"

/**
 * @description: 获取当前表中记录号为rid的记录
 * @param {Rid&} rid 记录号，指定记录的位置
 * @param {Context*} context
 * @return {unique_ptr<RmRecord>} rid对应的记录对象指针
 */
std::unique_ptr<RmRecord> RmFileHandle::get_record(const Rid& rid, Context* context) const {
    // Todo:
    // 1. 获取指定记录所在的page handle
    // 2. 初始化一个指向RmRecord的指针（赋值其内部的data和size）

    //加入行级锁
    {
        if(context!=nullptr){
            context->lock_mgr_->lock_shared_on_record(context->txn_,rid,fd_);
        }
    }
    int page_no = rid.page_no;
    auto page_handler = fetch_page_handle(page_no);

    int slot_no = rid.slot_no;
    char *data = page_handler.get_slot(slot_no);
    int size = file_hdr_.record_size;

    auto record = std::make_unique<RmRecord>(size, data);
    buffer_pool_manager_->unpin_page({fd_,page_no}, false);
    return record;
}

/**
 *
 * @param rid
 * @param page_buffer 缓存页
 * @param context
 * @return
 * @brief 注意，需要在函数外部，将最后一次fetch得到的page_buffer,unpin掉
 */
std::unique_ptr<RmRecord> RmFileHandle::get_record(const Rid& rid,RmPageHandle& page_buffer, Context* context) const {
    // Todo:
    // 1. 获取指定记录所在的page handle
    // 2. 初始化一个指向RmRecord的指针（赋值其内部的data和size）

    //加入行级锁
    {
        if(context!=nullptr){
            context->lock_mgr_->lock_shared_on_record(context->txn_,rid,fd_);
        }
    }
    int page_no = rid.page_no;
    if(page_buffer.page == nullptr){
        page_buffer = fetch_page_handle(page_no);
    }
    auto last_page_no = page_buffer.page->get_page_id().page_no;
    if(last_page_no != page_no){
        buffer_pool_manager_->unpin_page({fd_,last_page_no}, false);
        page_buffer = fetch_page_handle(page_no);
    }
    char *data = page_buffer.get_slot(rid.slot_no);

    return std::make_unique<RmRecord>(file_hdr_.record_size, data);
}

/**
 * @description: 在当前表中插入一条记录，不指定插入位置
 * @param {char*} buf 要插入的记录的数据
 * @param {Context*} context
 * @return {Rid} 插入的记录的记录号（位置）
 */
Rid RmFileHandle::insert_record(char* buf, Context* context) {
    // Todo:
    // 1. 获取当前未满的page handle
    // 2. 在page handle中找到空闲slot位置
    // 3. 将buf复制到空闲slot位置
    // 4. 更新page_handle.page_hdr中的数据结构
    // 注意考虑插入一条记录后页面已满的情况，需要更新file_hdr_.first_free_page_no

    // 得到一个未满的page handle
    auto page_handler = create_page_handle();
    // 找到空闲的slot
    int slot_index = Bitmap::first_bit(false, page_handler.bitmap, file_hdr_.num_records_per_page);//存在多次调用函数的问题，后续可以优化
    //int second_slot_index = Bitmap::first_bit(0,page_handler.bitmap + slot_index,file_hdr_.num_records_per_page);
    //加入行级锁

    {
        Rid r{page_handler.page->get_page_id().page_no,slot_index};
        //写日志
        //日志记录
        RmRecord rec(file_hdr_.record_size,buf);
        if(enable_logging && context != nullptr){
            //更新txn内的写集和写日
            std::unique_ptr<InsertLogRecord> insert_log(new InsertLogRecord(context->txn_->get_transaction_id(),context->txn_->get_prev_lsn(),
                                                                            rec,r,disk_manager_->get_file_name(fd_)));
            context->log_mgr_->add_log_to_buffer(insert_log.get());
            context->txn_->set_prev_lsn(insert_log->lsn_);
            page_handler.page->set_page_lsn(insert_log->lsn_);
        }
        //加入行级锁
        if(context!=nullptr){
            context->lock_mgr_->lock_exclusive_on_record(context->txn_,r,fd_);
        }
    }

    // 复制
    char *data = page_handler.get_slot(slot_index);
    memcpy(data, buf, file_hdr_.record_size);
    // 更新page_handle.page_hdr
    page_handler.page_hdr->num_records++;
    // 更新bitmap
    Bitmap::set(page_handler.bitmap, slot_index);
    auto page_id = page_handler.page->get_page_id();
    if (page_handler.page_hdr->num_records >= file_hdr_.num_records_per_page) {
        // 考虑这一页满的情况
        file_hdr_.first_free_page_no = page_handler.page_hdr->next_free_page_no;
        page_handler.page_hdr->next_free_page_no = -1;
        //及时更新header，利于故障终断后恢复
        //disk_manager_->write_page(fd_, RM_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
    }
    buffer_pool_manager_->unpin_page(page_id, true);
    return Rid{page_id.page_no, slot_index};
}


void RmFileHandle::load_record(const std::vector<std::unique_ptr<RmRecord>> &records,
                               std::vector<Rid>& rids,Context *context) {

    int rec_size = file_hdr_.record_size;
    int rec_nums = file_hdr_.num_records_per_page;
    //std::set<PageId> unpin_page_set;
    RmPageHandle page_buffer = create_page_handle();
    for(const auto &rec_iter : records){
        int free_slot_no = Bitmap::first_bit(false, page_buffer.bitmap, rec_nums);

        char* des = page_buffer.get_slot(free_slot_no);
        memcpy(des,rec_iter->data,rec_size);
        Bitmap::set(page_buffer.bitmap, free_slot_no);

        rids.emplace_back(Rid{page_buffer.page->get_page_id().page_no,free_slot_no});

        if(++page_buffer.page_hdr->num_records == file_hdr_.num_records_per_page){
            file_hdr_.first_free_page_no = page_buffer.page_hdr->next_free_page_no;
            buffer_pool_manager_->unpin_page(page_buffer.page->get_page_id(), true);
            //unpin_page_set.insert(page_buffer.page->get_page_id());
            page_buffer = create_page_handle();
        }
    }
    buffer_pool_manager_->unpin_page(page_buffer.page->get_page_id(),true);//小问题，可能最后一页刚刚创建，但没有数据写入
    //buffer_pool_manager_->unpin_pageSet(unpin_page_set,true);

}
/**
 * @description: 在当前表中的指定位置插入一条记录
 * @param {Rid&} rid 要插入记录的位置
 * @param {char*} buf 要插入记录的数据
 */
void RmFileHandle::insert_record(const Rid& rid, char* buf) {
    // TODO
    //是否需要锁

    // 找到指定的页
    int page_no = rid.page_no, slot_no = rid.slot_no;
    auto page_handler = fetch_page_handle(page_no);

    // 复制
    char *data = page_handler.get_slot(slot_no);
    memcpy(data, buf, file_hdr_.record_size);

    // 更新
    page_handler.page_hdr->num_records++;
    Bitmap::set(page_handler.bitmap, slot_no);
    if (page_handler.page_hdr->num_records >= file_hdr_.num_records_per_page) {
        // 考虑这一页满的情况
        file_hdr_.first_free_page_no = page_handler.page_hdr->next_free_page_no;
        page_handler.page_hdr->next_free_page_no = -1;
        //disk_manager_->write_page(fd_, RM_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
    }
    buffer_pool_manager_->unpin_page(page_handler.page->get_page_id(), true);
}

/**
 * @description: 删除记录文件中记录号为rid的记录
 * @param {Rid&} rid 要删除的记录的记录号（位置）
 * @param {Context*} context
 */
void RmFileHandle::delete_record(const Rid& rid, Context* context) {
    // Todo:
    // 1. 获取指定记录所在的page handle
    // 2. 更新page_handle.page_hdr中的数据结构
    // 注意考虑删除一条记录后页面未满的情况，需要调用release_page_handle()

    {
        if(context!=nullptr){
            context->lock_mgr_->lock_exclusive_on_record(context->txn_,rid,fd_);
        }
    }
    // 找到指定的页
    int page_no = rid.page_no, slot_no = rid.slot_no;
    auto page_handler = fetch_page_handle(page_no);

    // 删除
    memset(page_handler.get_slot(slot_no), 0, file_hdr_.record_size);//debug用

    Bitmap::reset(page_handler.bitmap, slot_no);
    if (page_handler.page_hdr->num_records == file_hdr_.num_records_per_page) {
        page_handler.page_hdr->num_records--;
        release_page_handle(page_handler);
    } else {
        page_handler.page_hdr->num_records--;
    }
    buffer_pool_manager_->unpin_page(page_handler.page->get_page_id(), true);
}


/**
 * @description: 更新记录文件中记录号为rid的记录
 * @param {Rid&} rid 要更新的记录的记录号（位置）
 * @param {char*} buf 新记录的数据
 * @param {Context*} context
 */
void RmFileHandle::update_record(const Rid& rid, char* buf, Context* context) {
    // Todo:
    // 1. 获取指定记录所在的page handle
    // 2. 更新记录

    if(context != nullptr){
        context->lock_mgr_->lock_exclusive_on_record(context->txn_,rid,fd_);
    }

    // 找到指定的页
    int page_no = rid.page_no, slot_no = rid.slot_no;
    auto page_handler = fetch_page_handle(page_no);
    memcpy(page_handler.get_slot(slot_no), buf, file_hdr_.record_size);
    buffer_pool_manager_->unpin_page(page_handler.page->get_page_id(), true);
}

/**
 * 以下函数为辅助函数，仅提供参考，可以选择完成如下函数，也可以删除如下函数，在单元测试中不涉及如下函数接口的直接调用
*/
/**
 * @description: 获取指定页面的页面句柄
 * @param {int} page_no 页面号
 * @return {RmPageHandle} 指定页面的句柄
 */
RmPageHandle RmFileHandle::fetch_page_handle(int page_no) const {
    // Todo:
    // 使用缓冲池获取指定页面，并生成page_handle返回给上层
    // if page_no is invalid, throw PageNotExistError exception

    if (page_no == INVALID_PAGE_ID) {
        //assert(0);
        throw PageNotExistError("", page_no);
    }

    /**
     * 获得file_hdr和页面指针
    */

    // 通过缓冲池得到页面
    // 构造PageId
    PageId page_id = {fd_, page_no};
    Page *page = buffer_pool_manager_->fetch_page(page_id);

    return RmPageHandle(&file_hdr_, page);
}

/**
 * @description: 创建一个新的page handle
 * @return {RmPageHandle} 新的PageHandle
 */
RmPageHandle RmFileHandle::create_new_page_handle() {
    // Todo:
    // 1.使用缓冲池来创建一个新page
    // 2.更新page handle中的相关信息
    // 3.更新file_hdr_

    // 创建一个新的page
    // 构造一个page_id
    PageId page_id;
    page_id.fd = fd_;
    Page *page = buffer_pool_manager_->new_page(&page_id);

    auto page_no = page_id.page_no;

    RmPageHandle page_handler(&file_hdr_, page);

    // 更新page_handler中的相关信息
    page_handler.page_hdr->next_free_page_no = -1;
    page_handler.page_hdr->num_records = 0;
    Bitmap::init(page_handler.bitmap, file_hdr_.bitmap_size);

    // 更新file_hdr_中的信息
    file_hdr_.num_pages++;
    file_hdr_.first_free_page_no = page_no;
    //disk_manager_->write_page(fd_, RM_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
    return page_handler;
}

/**
 * @brief 创建或获取一个空闲的page handle
 *
 * @return RmPageHandle 返回生成的空闲page handle
 * @note pin the page, remember to unpin it outside!
 */
RmPageHandle RmFileHandle::create_page_handle() {
    // Todo:
    // 1. 判断file_hdr_中是否还有空闲页
    //     1.1 没有空闲页：使用缓冲池来创建一个新page；可直接调用create_new_page_handle()
    //     1.2 有空闲页：直接获取第一个空闲页
    // 2. 生成page handle并返回给上层

    if (file_hdr_.first_free_page_no == INVALID_PAGE_ID) {
        // 没有空闲页
        return create_new_page_handle();
    }
    return fetch_page_handle(file_hdr_.first_free_page_no);
}

RmPageHandle RmFileHandle::create_buffer_page_handle() {
    // Todo:
    // 1. 判断file_hdr_中是否还有空闲页
    //     1.1 没有空闲页：使用缓冲池来创建一个新page；可直接调用create_new_page_handle()
    //     1.2 有空闲页：直接获取第一个空闲页
    // 2. 生成page handle并返回给上层

    PageId page_id;
    page_id.fd = fd_;
    Page *page = buffer_pool_manager_->new_page(&page_id);


    RmPageHandle page_handler(&file_hdr_, page);
    // 更新page_handler中的相关信息
    page_handler.page_hdr->next_free_page_no = -1;
    page_handler.page_hdr->num_records = 0;
    Bitmap::init(page_handler.bitmap, file_hdr_.bitmap_size);

    // 更新file_hdr_中的信息
    file_hdr_.num_pages++;
    //file_hdr_.first_free_page_no = page_no;//因为只有页满时才写入
    //disk_manager_->write_page(fd_, RM_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
    return page_handler;
}

/**
 * @description: 当一个页面从没有空闲空间的状态变为有空闲空间状态时，更新文件头和页头中空闲页面相关的元数据
 */
void RmFileHandle::release_page_handle(RmPageHandle&page_handle) {
    // Todo:
    // 当page从已满变成未满，考虑如何更新：
    // 1. page_handle.page_hdr->next_free_page_no
    // 2. file_hdr_.first_free_page_no
    page_handle.page_hdr->next_free_page_no = file_hdr_.first_free_page_no;
    file_hdr_.first_free_page_no = page_handle.page->get_page_id().page_no;
    //disk_manager_->write_page(fd_, RM_FILE_HDR_PAGE, (char *)&file_hdr_, sizeof(file_hdr_));
}