/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "sm_manager.h"

#include <sys/stat.h>
#include <unistd.h>

#include <fstream>
#include <future>

#include "index/ix.h"
#include "record/rm.h"
#include "record_printer.h"

#include "transaction/transaction.h"
#include "transaction/transaction_manager.h"
#include "thread_pool.h"
/**
 * @description: 判断是否为一个文件夹
 * @return {bool} 返回是否为一个文件夹
 * @param {string&} db_name 数据库文件名称，与文件夹同名
 */
bool SmManager::is_dir(const std::string& db_name) {
    struct stat st;
    return stat(db_name.c_str(), &st) == 0 && S_ISDIR(st.st_mode);
}


/**
 * @description: 创建数据库，所有的数据库相关文件都放在数据库同名文件夹下
 * @param {string&} db_name 数据库名称
 */
void SmManager::create_db(const std::string& db_name) {
    if (is_dir(db_name)) {
        throw DatabaseExistsError(db_name);
    }
    //为数据库创建一个子目录
    std::string cmd = "mkdir " + db_name;
    if (system(cmd.c_str()) < 0) {  // 创建一个名为db_name的目录
        throw UnixError();
    }
    if (chdir(db_name.c_str()) < 0) {  // 进入名为db_name的目录
        throw UnixError();
    }
    //创建系统目录
    DbMeta *new_db = new DbMeta();
    new_db->name_ = db_name;

    // 注意，此处ofstream会在当前目录创建(如果没有此文件先创建)和打开一个名为DB_META_NAME的文件
    std::ofstream ofs(DB_META_NAME);

    // 将new_db中的信息，按照定义好的operator<<操作符，写入到ofs打开的DB_META_NAME文件中
    ofs << *new_db;  // 注意：此处重载了操作符<<

    delete new_db;

    // 创建日志文件
    disk_manager_->create_file(LOG_FILE_NAME);

    // 回到根目录
    if (chdir("..") < 0) {
        throw UnixError();
    }
}

/**
 * @description: 删除数据库，同时需要清空相关文件以及数据库同名文件夹
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
void SmManager::drop_db(const std::string& db_name) {
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }
    std::string cmd = "rm -r " + db_name;
    if (system(cmd.c_str()) < 0) {
        throw UnixError();
    }
}

/**
 * @description: 打开数据库，找到数据库对应的文件夹，并加载数据库元数据和相关文件
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
void SmManager::open_db(const std::string& db_name) {
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }

    if (chdir(db_name.c_str()) < 0) {
        // 进入数据库文件夹
        throw UnixError();
    }

    // 加载数据库元数据
    std::ifstream ifs(DB_META_NAME);
    ifs >> db_;

    // 将表数据,index数据读入
    for (const auto &tab : db_.tabs_) {
        auto handle = rm_manager_->open_file(tab.second.name);
        fhs_[tab.second.name] = std::move(handle);
    }

}

/**
 * @description: 把数据库相关的元数据刷入磁盘中
 */

void SmManager::flush_meta() {
    // 默认清空文件
    std::ofstream ofs(DB_META_NAME);
    ofs << db_;
}

/**
 * @description: 关闭数据库并把数据落盘
 */
void SmManager::close_db() {
    flush_meta();
    for(auto &tab: db_.tabs_){
        buffer_pool_manager_->flush_all_pages(fhs_.at(tab.first)->GetFd());
        rm_manager_->close_file(fhs_.at(tab.first).get());
        fhs_.erase(tab.first);
        for(auto &index : tab.second.indexes){
            buffer_pool_manager_->flush_all_pages(ihs_.at(ix_manager_->get_index_name(tab.first, index.cols))->get_fd());
            ix_manager_->close_index(ihs_.at(ix_manager_->get_index_name(tab.first, index.cols)).get());
            ihs_.erase(ix_manager_->get_index_name(tab.first, index.cols));
        }
    }

    db_.name_.clear();
    db_.tabs_.clear();
    fhs_.clear();
    ihs_.clear();
//    flush_meta();
}

/**
 * @description: 显示所有的表,通过测试需要将其结果写入到output.txt,详情看题目文档
 * @param {Context*} context 
 */
void SmManager::show_tables(Context* context) {
    std::fstream outfile;
    outfile.open("output.txt", std::ios::out | std::ios::app);
    outfile << "| Tables |\n";
    RecordPrinter printer(1);
    printer.print_separator(context);
    printer.print_record({"Tables"}, context);
    printer.print_separator(context);
    for (auto &entry : db_.tabs_) {
        auto &tab = entry.second;
        printer.print_record({tab.name}, context);
        outfile << "| " << tab.name << " |\n";
    }
    printer.print_separator(context);
    outfile.close();
}

/**
 * @description: 显示表的元数据
 * @param {string&} tab_name 表名称
 * @param {Context*} context 
 */
void SmManager::desc_table(const std::string& tab_name, Context* context) {
    TabMeta &tab = db_.get_table(tab_name);

    std::vector<std::string> captions = {"Field", "Type", "Index"};
    RecordPrinter printer(captions.size());
    // Print header
    printer.print_separator(context);
    printer.print_record(captions, context);
    printer.print_separator(context);
    // Print fields
    for (auto &col : tab.cols) {
        std::vector<std::string> field_info = {col.name, coltype2str(col.type), col.index ? "YES" : "NO"};
        printer.print_record(field_info, context);
    }
    // Print footer
    printer.print_separator(context);
}

/**
 * @description: 创建表
 * @param {string&} tab_name 表的名称
 * @param {vector<ColDef>&} col_defs 表的字段
 * @param {Context*} context 
 */
void SmManager::create_table(const std::string& tab_name, const std::vector<ColDef>& col_defs, Context* context) {
    if (db_.is_table(tab_name)) {
        throw TableExistsError(tab_name);
    }
    // Create table meta
    int curr_offset = 0;
    TabMeta tab;
    tab.name = tab_name;
    tab.indexes.clear();
    tab.cols.clear();
    for (auto &col_def : col_defs) {
        ColMeta col = {.tab_name = tab_name,
                       .name = col_def.name,
                       .type = col_def.type,
                       .len = col_def.len,
                       .offset = curr_offset,
                       .index = false,
                       .is_agrregated = false};
        curr_offset += col_def.len;
        tab.cols.emplace_back(col);
    }
    // Create & open record file
    int record_size = curr_offset;  // record_size就是col meta所占的大小（表的元数据也是以记录的形式进行存储的）
    rm_manager_->create_file(tab_name, record_size);
    //throw DeBugError();
    db_.tabs_[tab_name] = tab;
    fhs_.emplace(tab_name, rm_manager_->open_file(tab_name));
    flush_meta();
    bug_flag = true;
}

/**
 * @description: 删除表
 * @param {string&} tab_name 表的名称
 * @param {Context*} context
 */
void SmManager::drop_table(const std::string& tab_name, Context* context) {
    // 判断表格是否存在
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }
    auto fd = disk_manager_->get_file_fd(tab_name);
    buffer_pool_manager_->delete_all_pages(fd);
    rm_manager_->close_file(fhs_[tab_name].get());
    rm_manager_->destroy_file(tab_name);
    fhs_.erase(tab_name);
    db_.tabs_.erase(tab_name);

    flush_meta();
    //bug_flag = true;
}

/**
 * @description: 创建索引
 * @param {string&} tab_name 表的名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
void SmManager::create_index(const std::string& tab_name, const std::vector<std::string>& col_names, Context* context) {
    // 判断表格是否存在
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }

    if(context != nullptr){
        context->lock_mgr_->lock_shared_on_table(context->txn_,fhs_[tab_name]->GetFd());
    }

    int col_num = col_names.size();
    int col_tot_len = 0;

    std::vector<ColMeta> index_cols;
    // 检查列是否存在
    TabMeta &tab_meta = db_.tabs_[tab_name];
    if (tab_meta.is_index(col_names)) {
        throw IndexExistsError(tab_name, col_names);
    }
    for (const auto &col : col_names) {
        if (!tab_meta.is_col(col)) {
            throw ColumnNotFoundError(col);
        }
        // 如果存在则将col_meta信息提取出来
        ColMeta col_meta = *(tab_meta.get_col(col));
        index_cols.push_back(col_meta);
        col_tot_len += col_meta.len;
    }

    IndexMeta index_meta;
    index_meta.tab_name = tab_name;
    index_meta.col_num = col_num;
    index_meta.col_tot_len = col_tot_len;
    index_meta.cols = index_cols;
    db_.tabs_[tab_name].indexes.push_back(std::move(index_meta));
    flush_meta();

    ix_manager_->create_index(tab_name, index_cols);
    auto index_name = ix_manager_->get_index_name(tab_name, col_names);
    ihs_[index_name] = ix_manager_->open_index(tab_name, col_names);

    // // 将所有原来的元素都插入这个索引中
    auto &index_handle = ihs_[index_name];
    auto &file_handle = fhs_[tab_name];
    RmScan rm_scan = RmScan(file_handle.get());
    char *key = new char[col_tot_len];
    while (!rm_scan.is_end()) {
        Rid rid = rm_scan.rid();
        auto record = file_handle->get_record(rid, context);
        int current_index = 0;
        for (const auto &col : col_names) {
            auto col_meta = tab_meta.get_col(col);
            int offset = col_meta->offset, len = col_meta->len;
            memcpy(key + current_index, record->data + offset, len);
            current_index += len;
        }
        index_handle->insert_entry(key, rid, (context== nullptr ? nullptr:context->txn_));
        rm_scan.next();
    }
    delete[] key;

}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
void SmManager::drop_index(const std::string& tab_name, const std::vector<std::string>& col_names, Context* context) {
    // 判断表格是否存在
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }
    //禁止符合最左匹配原则的写操作
    //禁止从ihs文件里读
    // TODO
    if(context != nullptr) {
        context->lock_mgr_->lock_shared_on_table(context->txn_, fhs_[tab_name]->GetFd());
    }

    TabMeta& tab_meta = db_.get_table(tab_name);
    
    if (!tab_meta.is_index(col_names)) {
        throw IndexNotFoundError(tab_name, col_names);
    }

    IndexMeta index_meta = *(tab_meta.get_index_meta(col_names));

    // ix_manager_->destroy_index(tab_name, col_names);
    for (auto i = tab_meta.indexes.begin(); i != tab_meta.indexes.end(); i++) {
        if (!same_index(index_meta, *i)) {
            continue;
        }

        auto index_name = ix_manager_->get_index_name(tab_name, col_names);

        //文件有可能处于未打开的状态
//        if(ihs_[index_name].get() == nullptr){
//            ihs_[index_name] = ix_manager_->open_index(tab_name, col_names);
//        }
        if(context != nullptr){
            context->lock_mgr_->lock_exclusive_on_table(context->txn_,ihs_[index_name]->get_fd());
        }

        //在删除大量数据前维护元数据，保证一致性
        tab_meta.indexes.erase(i);
        flush_meta();

        ix_manager_->close_index(ihs_[index_name].get());
        ix_manager_->destroy_index(tab_name, col_names);

        buffer_pool_manager_->delete_all_pages(ihs_[index_name]->get_fd());
        break;
    }
}

bool SmManager::same_index(const IndexMeta &meta1, const IndexMeta &meta2) {
    if (meta1.tab_name != meta2.tab_name) {
        return false;
    }

    if (meta1.col_tot_len != meta2.col_tot_len) {
        return false;
    }

    if (meta1.col_num != meta2.col_num) {
        return false;
    }

    for (int i = 0; i < meta1.col_num; i++) {
        if (meta1.cols[i].name != meta2.cols[i].name) {
            return false;
        }
    }

    return true;
}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<ColMeta>&} 索引包含的字段元数据
 * @param {Context*} context
 */
void SmManager::drop_index(const std::string& tab_name, const std::vector<ColMeta>& cols, Context* context) {
    std::vector<std::string> col_names;
    for (const auto &col : cols) {
        col_names.push_back(col.name);
    }

    drop_index(tab_name, col_names, context);
}

/**
 * @desciption: 展示索引
 * @param {string&} tab_name 表名称
 * @param {Context*} context
*/
void SmManager::show_index(const std::string& tab_name, Context* context) {
    std::fstream outfile;
    outfile.open("output.txt", std::ios::out | std::ios::app);
    RecordPrinter printer(3);

    printer.print_separator(context);

    std::vector<IndexMeta> index_metas = db_.tabs_[tab_name].indexes;
    
    for (const auto &meta : index_metas) {
        std::vector<std::string> rec_str;
        rec_str.push_back(tab_name);
        rec_str.push_back("unique");
        std::string final_col = "(";

        int size = meta.col_num;
        for (int i = 0; i < size - 1; i++) {
            final_col += meta.cols[i].name;
            final_col += ",";
        }

        final_col += meta.cols[size - 1].name;
        final_col += ")";
        rec_str.push_back(final_col);
        printer.print_record(std::move(rec_str), context);
        outfile << "| " << tab_name << " | " << "unique" << " | " << final_col << " |" << std::endl;
    }

    printer.print_separator(context);
    outfile.close();
}


void SmManager::recovery_create_index(const std::string &tab_name, const std::vector<std::string> &col_names){

    int col_num = col_names.size();
    int col_tot_len = 0;

    std::vector<ColMeta> index_cols;
    // 检查列是否存在
    TabMeta &tab_meta = db_.tabs_[tab_name];
    assert(tab_meta.is_index(col_names));

    for (const auto &col : col_names) {
        if (!tab_meta.is_col(col)) {
            assert(0);
            throw ColumnNotFoundError(col);
        }
        // 如果存在则将col_meta信息提取出来
        ColMeta col_meta = *(tab_meta.get_col(col));
        col_meta.index = true;
        index_cols.push_back(col_meta);
        col_tot_len += col_meta.len;
    }

    IndexMeta index_meta;
    index_meta.tab_name = tab_name;
    index_meta.col_num = col_num;
    index_meta.col_tot_len = col_tot_len;
    index_meta.cols = index_cols;
//    db_.tabs_[tab_name].indexes.push_back(std::move(index_meta));
//    flush_meta();
    auto index_name = ix_manager_->get_index_name(tab_name,col_names);//获得文件名
    if(disk_manager_->is_file(index_name)){
        //还没有删除过
        disk_manager_->destroy_file(index_name);
    }

    ix_manager_->create_index(tab_name, index_cols);
    assert(ihs_.count(index_name) == 0);
    ihs_[index_name] = ix_manager_->open_index(tab_name, col_names);

    // // 将所有原来的元素都插入这个索引中
    auto &index_handle = ihs_[index_name];
    auto &file_handle = fhs_[tab_name];
    RmScan rm_scan = RmScan(file_handle.get());
    while (!rm_scan.is_end()) {
        Rid rid = rm_scan.rid();
        auto record = file_handle->get_record(rid, nullptr);
        char *key = new char[col_tot_len];
        int current_index = 0;
        for (const auto &col : col_names) {
            auto col_meta = tab_meta.get_col(col);
            int offset = col_meta->offset, len = col_meta->len;
            memcpy(key + current_index, record->data + offset, len);
            current_index += len;
        }
        try{
            index_handle->insert_entry(key, rid, nullptr);
        }catch(InternalError &err){
            assert(0);
            auto fh_ = fhs_.at(tab_name).get();
            fh_->delete_record(rm_scan.rid(),nullptr);
        }
        rm_scan.next();
    }

}





void SmManager::load_data(std::string file_name, std::string tab_name,LockManager* lock_mgr,Transaction* txn) {

    auto tab_meta = this->db_.tabs_.at(tab_name);

    std::ifstream ifs(file_name);
    std::string s;
    std::getline(ifs,s);//跳过第一行
    if(s.empty())assert(0);

    s.clear();
    RmFileHandle* fh_ = this->fhs_.at(tab_name).get();
    RmPageHandle page_buffer = fh_->create_page_handle();//保证单线程
    std::vector<Value> row(tab_meta.cols.size());
    auto record_size = fh_->get_file_hdr().record_size;
    std::vector<std::unique_ptr<RmRecord>> records;
    std::vector<Rid> rids;

    while(std::getline(ifs, s)){
        auto rec = std::make_unique<RmRecord>(record_size);
        size_t start = 0;
        int col_index = 0;
        auto line_len =s.size();
        const char *p = s.data();
        for(size_t end = 0; end <= line_len; ++end){
            if (end == line_len || *(p + end) == ',') {
                auto &col_meta = tab_meta.cols.at(col_index);
                char *dest = rec->data + col_meta.offset;
                auto len = end - start;
                switch (col_meta.type) {
                    case ColType::TYPE_INT:{
                        std::string v(p + start, len);
                        *(int *)dest = std::stoi(v);
                        break;
                    }
                    case ColType::TYPE_FLOAT:{
                        std::string v(p + start,len);
                        *(float *)dest = std::stof(v);
                        break;
                    }
                    case ColType::TYPE_STRING:{
                        memcpy(dest, p + start, len);
                        if(len != col_meta.len){
                            memset(dest + len, 0, col_meta.len - len);
                        }
                        break;
                    }
                    default:
                        assert(0);
                }
                start = end + 1;
                ++col_index;
            }
        }
        records.push_back(std::move(rec));
    }
    fh_->load_record(records,rids, nullptr);
    buffer_pool_manager_->unpin_page(page_buffer.page->get_page_id(),true);

    auto rec_num = records.size();
    std::vector<PageId> unpin_index_page;
    for(auto &index : tab_meta.indexes) {
        auto ih = ihs_.at(get_ix_manager()->get_index_name(tab_meta.name, index.cols)).get();
        char *key = new char[index.col_tot_len];
        for(size_t i = 0; i < rec_num; i++) {
            int offset = 0;
            for(int j = 0; j < index.col_num; j++) {
                memcpy(key + offset, records[i]->data + index.cols[j].offset, index.cols[j].len);
                offset += index.cols[j].len;
            }
            ih->load_insert_entry(key, rids[i],unpin_index_page);//假设单线程
        }
        delete[] key;
    }
    ifs.close();

}

void SmManager::async_load_data(const std::string &file_name, const std::string &tab_name, Context *context,TransactionManager * txn_mgr) {
    active_tasks_++;
    load_threads.emplace_back(&SmManager::load_data, this, file_name, tab_name, nullptr, nullptr);
    if(active_tasks_ == 9){
        //若要detach模式，需要修改wait-die原则
        for (std::thread & thread : load_threads) {
            thread.join();
        }
        assert(0);
    }
}
