/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include <algorithm>
#include "lock_manager.h"

/**
 * @description: 申请行级共享锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID 记录所在的表的fd
 * @param {int} tab_fd
 */
bool LockManager::lock_shared_on_record(Transaction* txn, const Rid& rid, int tab_fd) {
    checkIsolationLevel(txn,LockMode::SHARED);
    LockDataId lock_id(tab_fd,rid,LockDataType::RECORD);

    std::unique_lock<std::mutex> l(latch_);

    //若对应id没有队列，但马上插入请求
    auto lock_queue = lock_table_.find(lock_id);
    if(lock_queue == lock_table_.end()){
        LockRequest request(txn->get_transaction_id(),LockMode::SHARED);
        txn->set_state((TransactionState::GROWING));
        txn->get_lock_set()->insert(lock_id);
        std::unique_lock<std::mutex> q_lock(lock_table_[lock_id].latch_);
        request.granted_ = true;
        lock_table_[lock_id].request_queue_.push_back(request);
        lock_table_[lock_id].group_lock_mode_ = GroupLockMode::S;
        return true;
    }

    //对应id有队列
    auto request_queue = &lock_queue->second;
    //获得对应队列的锁，用于并发
    std::unique_lock<std::mutex> q_lock(request_queue->latch_);
    l.unlock();

    bool need_upgrade = dealOnSameSource(txn,request_queue,LockMode::SHARED,q_lock);
    if(txn->get_state() == TransactionState::ABORTED)return false;
    if(!need_upgrade)txn->get_lock_set()->emplace(lock_id);
    txn->set_state((TransactionState::GROWING));
    return true;
}

/**
 * @description: 申请行级排他锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID
 * @param {int} tab_fd 记录所在的表的fd
 */
bool LockManager::lock_exclusive_on_record(Transaction* txn, const Rid& rid, int tab_fd) {
    checkIsolationLevel(txn,LockMode::EXLUCSIVE);
    LockDataId lock_id(tab_fd,rid,LockDataType::RECORD);
    std::unique_lock<std::mutex> l(latch_);
    //若对应id没有队列，但马上插入请求
    auto lock_queue = lock_table_.find(lock_id);
    if(lock_queue == lock_table_.end()){
        LockRequest request(txn->get_transaction_id(),LockMode::EXLUCSIVE);
        txn->set_state((TransactionState::GROWING));
        txn->get_lock_set()->emplace(lock_id);
        std::unique_lock<std::mutex> q_lock(lock_table_[lock_id].latch_);
        request.granted_ = true;
        lock_table_[lock_id].request_queue_.push_back(request);
        lock_table_[lock_id].group_lock_mode_ = GroupLockMode::X;
        return true;
    }

    //对应id有队列
    auto request_queue = &lock_queue->second;
    //获得对应队列的锁，用于并发
    std::unique_lock<std::mutex> q_lock(request_queue->latch_);
    l.unlock();

    bool need_upgrade = dealOnSameSource(txn,request_queue,LockMode::EXLUCSIVE,q_lock);
    if(txn->get_state() == TransactionState::ABORTED)return false;
    if(!need_upgrade)txn->get_lock_set()->emplace(lock_id);
    txn->set_state((TransactionState::GROWING));
    return true;
}

/**
 * @description: 申请表级读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_shared_on_table(Transaction* txn, int tab_fd) {
    checkIsolationLevel(txn,LockMode::SHARED);
    LockDataId lock_id(tab_fd,LockDataType::TABLE);

    std::unique_lock<std::mutex> l(latch_);

    //若对应id没有队列，但马上插入请求
    auto lock_queue = lock_table_.find(lock_id);
    if(lock_queue == lock_table_.end()){
        LockRequest request(txn->get_transaction_id(),LockMode::SHARED);
        txn->set_state((TransactionState::GROWING));
        txn->get_lock_set()->emplace(lock_id);
        std::unique_lock<std::mutex> q_lock(lock_table_[lock_id].latch_);
        request.granted_ = true;
        lock_table_[lock_id].request_queue_.push_back(request);
        lock_table_[lock_id].group_lock_mode_ = GroupLockMode::S;
        return true;
    }

    //对应id有队列
    auto request_queue = &lock_queue->second;
    //获得对应队列的锁，用于并发
    std::unique_lock<std::mutex> q_lock(request_queue->latch_);
    l.unlock();

    bool need_upgrade = dealOnSameSource(txn,request_queue,LockMode::SHARED,q_lock);
    if(txn->get_state() == TransactionState::ABORTED)return false;
    if(!need_upgrade)txn->get_lock_set()->emplace(lock_id);
    txn->set_state((TransactionState::GROWING));
    return true;
}

/**
 * @description: 申请表级写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_exclusive_on_table(Transaction* txn, int tab_fd) {
    checkIsolationLevel(txn,LockMode::EXLUCSIVE);
    LockDataId lock_id(tab_fd,LockDataType::TABLE);

    std::unique_lock<std::mutex> l(latch_);

    //若对应id没有队列，但马上插入请求
    auto lock_queue = lock_table_.find(lock_id);
    if(lock_queue == lock_table_.end()){
        LockRequest request(txn->get_transaction_id(),LockMode::EXLUCSIVE);
        txn->set_state((TransactionState::GROWING));
        txn->get_lock_set()->emplace(lock_id);
        std::unique_lock<std::mutex> q_lock(lock_table_[lock_id].latch_);
        request.granted_ = true;
        lock_table_[lock_id].request_queue_.push_back(request);
        lock_table_[lock_id].group_lock_mode_ = GroupLockMode::X;
        return true;
    }

    //对应id有队列
    auto request_queue = &lock_queue->second;
    //获得对应队列的锁，用于并发
    std::unique_lock<std::mutex> q_lock(request_queue->latch_);
    l.unlock();
    bool need_upgrade = dealOnSameSource(txn,request_queue,LockMode::EXLUCSIVE,q_lock);
    if(txn->get_state() == TransactionState::ABORTED)return false;
    if(!need_upgrade)txn->get_lock_set()->emplace(lock_id);
    txn->set_state((TransactionState::GROWING));
    return true;
}

bool LockManager::lock_exclusive_on_table(Transaction* txn, int tab_fd,bool is_supper){
    checkIsolationLevel(txn,LockMode::EXLUCSIVE);
    LockDataId lock_id(tab_fd,LockDataType::TABLE);

    std::unique_lock<std::mutex> l(latch_);

    //若对应id没有队列，但马上插入请求
    auto lock_queue = lock_table_.find(lock_id);
    if(lock_queue == lock_table_.end()){
        LockRequest request(txn->get_transaction_id(),LockMode::EXLUCSIVE);
        txn->set_state((TransactionState::GROWING));
        txn->get_lock_set()->emplace(lock_id);
        std::unique_lock<std::mutex> q_lock(lock_table_[lock_id].latch_);
        request.granted_ = true;
        request.is_supper_ = is_supper;
        lock_table_[lock_id].request_queue_.push_back(request);
        return true;
    }

    //对应id有队列
    auto request_queue = &lock_queue->second;
    //获得对应队列的锁，用于并发
    std::unique_lock<std::mutex> q_lock(request_queue->latch_);
    l.unlock();
    bool need_upgrade = dealOnSameSource(txn,request_queue,LockMode::EXLUCSIVE,q_lock);
    if(txn->get_state() == TransactionState::ABORTED)return false;
    if(!need_upgrade)txn->get_lock_set()->emplace(lock_id);
    txn->set_state((TransactionState::GROWING));
    return true;
}


/**
 * @description: 申请表级意向读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IS_on_table(Transaction* txn, int tab_fd) {
    checkIsolationLevel(txn,LockMode::INTENTION_SHARED);
    LockDataId lock_id(tab_fd,LockDataType::TABLE);

    std::unique_lock<std::mutex> l(latch_);

    //若对应id没有队列，但马上插入请求
    auto lock_queue = lock_table_.find(lock_id);
    if(lock_queue == lock_table_.end()){
        LockRequest request(txn->get_transaction_id(),LockMode::INTENTION_SHARED);
        txn->set_state((TransactionState::GROWING));
        txn->get_lock_set()->emplace(lock_id);
        std::unique_lock<std::mutex> q_lock(lock_table_[lock_id].latch_);
        request.granted_ = true;
        lock_table_[lock_id].request_queue_.push_back(request);
        lock_table_[lock_id].group_lock_mode_ = GroupLockMode::IS;
        return true;
    }

    //对应id有队列
    auto request_queue = &lock_queue->second;
    //获得对应队列的锁，用于并发
    std::unique_lock<std::mutex> q_lock(request_queue->latch_);
    l.unlock();

    bool need_upgrade = dealOnSameSource(txn,request_queue,LockMode::INTENTION_SHARED,q_lock);
    if (txn->get_state() == TransactionState::ABORTED)return false;
    if(!need_upgrade)txn->get_lock_set()->emplace(lock_id);
    txn->set_state((TransactionState::GROWING));
    return true;
}

/**
 * @description: 申请表级意向写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IX_on_table(Transaction* txn, int tab_fd) {
    checkIsolationLevel(txn,LockMode::INTENTION_EXCLUSIVE);
    LockDataId lock_id(tab_fd,LockDataType::TABLE);

    std::unique_lock<std::mutex> l(latch_);

    //若对应Did没有队列，马上插入请求
    //表示第一次申请这个锁
    auto lock_queue = lock_table_.find(lock_id);
    if(lock_queue == lock_table_.end()){
        LockRequest request(txn->get_transaction_id(),LockMode::INTENTION_EXCLUSIVE);
        txn->set_state((TransactionState::GROWING));
        txn->get_lock_set()->emplace(lock_id);
        std::unique_lock<std::mutex> q_lock(lock_table_[lock_id].latch_);
        request.granted_ = true;
        lock_table_[lock_id].request_queue_.push_back(request);
        lock_table_[lock_id].group_lock_mode_ = GroupLockMode::IX;
        return true;
    }

    //资源已经被占有，或者有其他事务申请
    auto request_queue = &lock_queue->second;
    //获得对应队列的锁，用于并发
    std::unique_lock<std::mutex> q_lock(request_queue->latch_);
    l.unlock();

    bool need_upgrade = dealOnSameSource(txn,request_queue,LockMode::INTENTION_EXCLUSIVE,q_lock);
    if(txn->get_state() == TransactionState::ABORTED)return false;
    if(!need_upgrade)txn->get_lock_set()->emplace(lock_id);
    txn->set_state((TransactionState::GROWING));
    return true;
}

/**
 * @description 锁住rid的前一个空
 * @param txn
 * @param index_fd
 * @param rid ，slot可能为0，表示新插入末尾，需等待
 * @return
 */
bool LockManager::lock_gap_on_index(Transaction *txn, int index_fd, const Rid &rid) {
    checkIsolationLevel(txn,LockMode::SHARED);
    LockDataId lock_id(index_fd,rid,LockDataType::GAP);
    std::unique_lock<std::mutex> l(latch_);

    //若对应Did没有队列，马上插入请求
    //表示第一次申请这个锁
    auto lock_queue = lock_table_.find(lock_id);
    if(lock_queue == lock_table_.end()){
        LockRequest request(txn->get_transaction_id(),LockMode::SHARED);
        txn->set_state((TransactionState::GROWING));
        txn->get_lock_set()->emplace(lock_id);
        std::unique_lock<std::mutex> q_lock(lock_table_[lock_id].latch_);
        request.granted_ = true;
        lock_table_[lock_id].request_queue_.push_back(request);
        return true;
    }
    //资源已经被占有，或者有其他事务申请
    auto request_queue = &lock_queue->second;
    //获得对应队列的锁，用于并发
    std::unique_lock<std::mutex> q_lock(request_queue->latch_);
    l.unlock();
    auto iter = request_queue->request_queue_.begin();
    //若是同一事务，获得间隙锁，则不用阻塞
    for(;iter != request_queue->request_queue_.end();++iter){
        if(iter->granted_ && iter->txn_id_ == txn->get_transaction_id())return true;
    }
    LockRequest lock_request(txn->get_transaction_id(), LockMode::SHARED);
    auto lock_request_iter = request_queue->request_queue_.insert(request_queue->request_queue_.end(),lock_request);
//    request_queue->cv_.wait(q_lock, [&request_queue, &lock_request_iter, &txn, this]{
//        // wait die
//        if(txn->get_state()==TransactionState::ABORTED){
//            return true;
//        }
//        txn_id_t high_priotity_id = txn->get_transaction_id();
//        if(request_queue->request_queue_.size() == 1)return true;//无人争抢
//        for(auto item : request_queue->request_queue_){
//            if(item.granted_)return false;
//            if(item.txn_id_ < txn->get_transaction_id()){
//                high_priotity_id = item.txn_id_;
////                txn->set_state(TransactionState::ABORTED);
////                return true;
//            }
//        }
//        if(high_priotity_id < txn->get_transaction_id()){
//            txn->set_state(TransactionState::ABORTED);
//            return true;
//        }
//        if(high_priotity_id == txn->get_transaction_id()){
//            return true;
//        }
//        return false;
//    });
//    if(txn->get_state() == TransactionState::ABORTED){
//        request_queue->request_queue_.erase(lock_request_iter);
//        throw TransactionAbortException (txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
//    }
    lock_request_iter->granted_ = true;
    txn->get_lock_set()->emplace(lock_id);
    txn->set_state((TransactionState::GROWING));
    return true;
}

/**
 * @description 查看对应的rid的间隙锁是否被获取
 */
bool LockManager::try_lock_gap_on_index(Transaction *txn, int index_fd, const Rid &rid) {
    checkIsolationLevel(txn,LockMode::SHARED);
    LockDataId lock_id(index_fd,rid,LockDataType::GAP);
    std::unique_lock<std::mutex> l(latch_);

    //若对应Did没有队列，马上插入请求
    //表示第一次申请这个锁
    auto lock_queue = lock_table_.find(lock_id);
    if(lock_queue == lock_table_.end()){
        LockRequest request(txn->get_transaction_id(),LockMode::SHARED);
        txn->set_state((TransactionState::GROWING));
        txn->get_lock_set()->emplace(lock_id);
        std::unique_lock<std::mutex> q_lock(lock_table_[lock_id].latch_);
        request.granted_ = true;
        lock_table_[lock_id].request_queue_.push_back(request);
        return true;
    }
    //资源已经被占有，或者有其他事务申请
    auto request_queue = &lock_queue->second;
    //获得对应队列的锁，用于并发
    std::unique_lock<std::mutex> q_lock(request_queue->latch_);
    l.unlock();
    auto iter = request_queue->request_queue_.begin();
    //若是同一事务，获得间隙锁，则不用阻塞
    for(;iter != request_queue->request_queue_.end();++iter){
        if(iter->granted_ && iter->txn_id_ == txn->get_transaction_id()){
            return true;
        }
    }
    //放入队列阻塞
    LockRequest lock_request(txn->get_transaction_id(), LockMode::SHARED);
    auto lock_request_iter = request_queue->request_queue_.insert(request_queue->request_queue_.end(),lock_request);
    request_queue->cv_.wait(q_lock, [&request_queue, &lock_request_iter, &txn, this]{
        // wait die
        if(txn->get_state()==TransactionState::ABORTED){
            return true;
        }
        txn_id_t high_prioritity_id = txn->get_transaction_id();
        for(auto item : request_queue->request_queue_){
            if(item.granted_ && item.txn_id_ > txn->get_transaction_id())return false;
            high_prioritity_id = (item.txn_id_ < txn->get_transaction_id())?item.txn_id_ : high_prioritity_id;
        }
        if(high_prioritity_id < txn->get_transaction_id()){
            txn->set_state(TransactionState::ABORTED);
            return true;
        }
        return high_prioritity_id == txn->get_transaction_id();
    });
    if(txn->get_state() == TransactionState::ABORTED){
        request_queue->request_queue_.erase(lock_request_iter);
        throw TransactionAbortException (txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    //request_queue->request_queue_.erase(lock_request_iter);
    lock_request_iter->granted_ = true;
    txn->get_lock_set()->emplace(lock_id);
    txn->set_state((TransactionState::GROWING));
    return true;
}

bool LockManager::try_erase_gap_on_index(Transaction *txn, int index_fd, const Rid &rid) {
    LockDataId lock_id(index_fd,rid,LockDataType::GAP);
    std::unique_lock<std::mutex> l(latch_);
    //若对应Did没有队列，马上插入请求
    //表示第一次申请这个锁
    auto lock_queue = lock_table_.find(lock_id);
    if(lock_queue == lock_table_.end()){
        return true;
    }
    //资源已经被占有，
    auto request_queue = &lock_queue->second;
    //获得对应队列的锁，用于并发
    std::unique_lock<std::mutex> q_lock(request_queue->latch_);
    l.unlock();
    auto iter = request_queue->request_queue_.begin();
    //只删除同一事务的间隙锁
    for(;iter != request_queue->request_queue_.end();++iter){
        if(iter->granted_ && iter->txn_id_ == txn->get_transaction_id()){
            request_queue->request_queue_.erase(iter);
            break;
        }
    }
    //这里没有维护锁集lock_set
    return true;
}

/**
 * @description: 释放锁
 * @return {bool} 返回解锁是否成功
 * @param {Transaction*} txn 要释放锁的事务对象指针
 * @param {LockDataId} lock_data_id 要释放的锁ID
 */
bool LockManager::unlock(Transaction* txn, LockDataId lock_data_id) {

    std::unique_lock<std::mutex> l(latch_);
    auto request_queue = &lock_table_[lock_data_id];
    std::unique_lock<std::mutex> l_q(request_queue->latch_);
    l.unlock();
    txn->set_state(TransactionState::SHRINKING);
    auto iter = request_queue->request_queue_.begin();
    while(iter != request_queue->request_queue_.end()){
        if(iter->txn_id_ == txn->get_transaction_id()){
            request_queue->request_queue_.erase(iter);
            request_queue->cv_.notify_all();
            break;
            //continue;
        }
        iter++;
    }
    if (request_queue->request_queue_.empty()) {
        lock_table_.erase(lock_data_id);
    }
    return true;
}

/**
 *help function
 */

/**
 *
 * @param txn
 * @param lock_mode
 * @return 是否符合隔离级别
 */
bool LockManager::checkIsolationLevel(Transaction *txn, LockManager::LockMode lock_mode) {
    if (txn->get_state() == TransactionState::DEFAULT) {
        txn->set_state(TransactionState::GROWING);
    }

    switch (txn->get_isolation_level()) {
        case IsolationLevel::READ_UNCOMMITTED:
            if (lock_mode != LockMode::EXLUCSIVE && lock_mode != LockMode::INTENTION_EXCLUSIVE) {
                // Attempted an unsupported lock mode under READ_UNCOMMITTED
                txn->set_state(TransactionState::ABORTED);
                throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_SHARED_ON_READ_UNCOMMITTED);
            }
            break;

        case IsolationLevel::READ_COMMITTED:
            if (txn->get_state() == TransactionState::SHRINKING) {
                if (lock_mode != LockMode::INTENTION_SHARED && lock_mode != LockMode::SHARED) {
                    txn->set_state(TransactionState::ABORTED);
                    throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
                }
            }
            break;

        case IsolationLevel::REPEATABLE_READ:
            if (txn->get_state() == TransactionState::SHRINKING) {
                // In REPEATABLE_READ, no new locks can be acquired during the shrinking phase
                txn->set_state(TransactionState::ABORTED);
                throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
            }
            break;

        case IsolationLevel::SERIALIZABLE:
            if (txn->get_state() == TransactionState::SHRINKING) {
                // In SERIALIZABLE, no new locks can be acquired during the shrinking phase
                txn->set_state(TransactionState::ABORTED);
                throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
            }
            break;

        default:
            txn->set_state(TransactionState::ABORTED);
    }

    return true;
}

/**
 *
 * @param old_mode 原先持有的锁
 * @param new_mode 所需要添加的锁；若可以更新，则为升级后的锁类型
 * @return true 如果可以更新
 */
bool LockManager::isCompatibleModeUpgrade(LockManager::LockMode old_mode, LockManager::LockMode& new_mode) {
    switch (old_mode) {
        case LockMode::INTENTION_SHARED:
            //assert(old_mode != LockMode::INTENTION_SHARED);
            return new_mode == LockMode::SHARED || new_mode == LockMode::EXLUCSIVE || new_mode == LockMode::S_IX || new_mode == LockMode::INTENTION_EXCLUSIVE;
        //可以变为SIX
        case LockMode::SHARED:
            if(new_mode == LockMode::INTENTION_EXCLUSIVE){
                new_mode = LockMode::S_IX;
                return true;
            }
            return new_mode == LockMode::EXLUCSIVE;
        case LockMode::INTENTION_EXCLUSIVE:
            if(new_mode == LockMode::SHARED){
                new_mode = LockMode::S_IX;
                return true;
            }
            return new_mode == LockMode::EXLUCSIVE;
        case LockMode::S_IX:
            return new_mode == LockMode::EXLUCSIVE;
        default:
            break;
    }
    return false;
}

/**
 *
 * @param mode1 原来的锁
 * @param mode2 要加的锁
 * @return
 */
bool LockManager::isCompatibleLockMode(LockManager::LockMode mode1, LockManager::LockMode mode2) {
    switch (mode2) {
        case LockMode::INTENTION_SHARED:
            return mode1 != LockMode::EXLUCSIVE;
        case LockMode::SHARED:
            return mode1 == LockMode::INTENTION_SHARED || mode1 == LockMode::SHARED;
        case LockMode::INTENTION_EXCLUSIVE:
            //return mode1 == LockMode::INTENTION_SHARED || mode1 == LockMode::INTENTION_EXCLUSIVE || mode1 == LockMode::SHARED;
            return mode1 == LockMode::INTENTION_SHARED || mode1 == LockMode::INTENTION_EXCLUSIVE;
        case LockMode::S_IX:
            return mode1 == LockMode::S_IX;
        case LockMode::EXLUCSIVE:
            return false;
        default:
            //assert(0);
            return false;
            break;
    }
    return true;
}

/**
 *@brief 不对同一tid进行判断，默认true
 * @param request_queue
 * @param request
 * @param occupied_id 现在占有资源上锁的id,或者正在等待获取锁的、优先级最高的txn对应的id
 * @return bool,表示不同的tid与当前队列内的granted的lockmode是否兼容；或者全部granted；或者非granted的tid一样
 */
bool LockManager::checkLockToQueueCompatible(LockManager::LockRequestQueue *request_queue,
                                             LockManager::LockRequest *request,txn_id_t &occupied_id) {

    bool res = true;
    bool is_supper = false;
    for(auto iter : request_queue->request_queue_){
        if(!iter.granted_)continue;
        //相同的判断是否可升级，不同则判断是否冲突

        if(iter.txn_id_ != request->txn_id_ && !isCompatibleLockMode(iter.lock_mode_,request->lock_mode_)){
            res = false;
            if(iter.txn_id_ < occupied_id){
                occupied_id = iter.txn_id_;
                is_supper = iter.is_supper_;
            }
        }
    }
    if(is_supper){
        occupied_id = std::numeric_limits<int32_t>::max();//txn_id最小，在wait-die策略下，所有人都需要等待
    }

    return res;

}

/**
 *
 * @param txn
 * @param request_queue
 * @param lock_mode
 * @param q_lock
 * @return true if upgrade
 */
bool LockManager::dealOnSameSource(Transaction *txn, LockManager::LockRequestQueue *request_queue,
                                   LockManager::LockMode lock_mode, std::unique_lock<std::mutex> &q_lock) {
    //TODO
    //input:txn,遇到的锁情况：其他事务占有，同一事务占有

    //队列内遇到的情况：
    //case1:都没有事务占有，即没有事务granted；case2:不同tid占有；case3：同一tid占有且队列内只有这一id；case4：不同tid和同一tid都占有；
    //case1：直接放入队列等待，对于非同tid的事务，采用wait-die判断
    //case2：与case1操作相同
    //case3：检查是否更新，若不更新，放入队列内等待；更新，则改变lock_mode
    //case4: 判断是否兼容(这里只与不同tid的granted事务比较）
    //case4.1:若兼容,判断是否更新，若不更新，放入队列内等待；
    //case4.2 若不兼容，放入等待队列



    bool need_upgrade = false;
    bool exist_same_tid = false;
    auto txn_id = txn->get_transaction_id();
    auto it = request_queue->request_queue_.begin();
    for(;it != request_queue->request_queue_.end();++it){
        if(it->txn_id_ != txn->get_transaction_id())continue;//可能导致case1，2没有判断
        exist_same_tid = true;
        //判断是否需要在同一事务上升级
        if(lock_mode == it->lock_mode_)return false;//同种类型锁，不处理
        need_upgrade = isCompatibleModeUpgrade(it->lock_mode_,lock_mode);//需保证这是最高级别的锁
        if(!need_upgrade && it->granted_){
            // 已经获取更高级别的锁(大锁), 上锁成功
            return false;
        }

        //竞争锁，为了提高效率，判断是否要升级
        //不升级说明前者的级别更高，这里直接释放
        if(!need_upgrade && !it->granted_){
            request_queue->cv_.wait(q_lock,[&request_queue,&it,&txn,&lock_mode,this]{
                if(txn->get_state() == TransactionState::ABORTED)return true;
                txn_id_t occupied_id = txn->get_transaction_id();
                if(checkLockToQueueCompatible(request_queue,&(*it),occupied_id)){
                    //兼容，
                    bool need_upgrade = isCompatibleModeUpgrade(it->lock_mode_,lock_mode);
                    if(need_upgrade)return !request_queue->upgrading_;
                    //不用升级锁
                    return true;
                }
                //wait die策略，与其他锁冲突
                //相等？
                if(occupied_id < txn->get_transaction_id()){
                    txn->set_state(TransactionState::ABORTED);
                    //回收txn，aborted
                    return true;
                }
                //优先级更高，继续等待或者id相同，不与同一事务上的其他并行锁相容
                return false;
            });
        }
        //可以更新
        //一次只能有一个更新
        if(request_queue->upgrading_)throw TransactionAbortException(txn_id,AbortReason::UPGRADE_CONFLICT);

        //准备更新
        //检查与其他锁是否冲突，冲突则wait-die
        //或者与本事务的其他锁冲
        request_queue->upgrading_ = true;
        it->lock_mode_ = lock_mode;//直接改的原因，因为采用2pl，所以待升级的锁会一直在队列内，我这里强制他一定要升级，可能降低并发量
        it->granted_ = false;
        request_queue->cv_.wait(q_lock,[&request_queue,&it,&txn,&lock_mode,this]{
            if(txn->get_state() == TransactionState::ABORTED)return true;
            txn_id_t occupied_id = txn->get_transaction_id();
            //队列兼容
            if(checkLockToQueueCompatible(request_queue,&(*it),occupied_id)){
                //bool need_upgrade = isCompatibleModeUpgrade(it->lock_mode_,lock_mode);
                //if(need_upgrade)return !request_queue->upgrading_;
                //不用升级锁
                return true;
            }

            //不兼容，等待释放冲突锁
            //这里使用wait die策略
            if(occupied_id < txn->get_transaction_id()){
                txn->set_state(TransactionState::ABORTED);
                //回收txn，aborted
                return true;
            }
            //优先级更高，继续等待或者id相同，不与同一事务上的其他并行锁相容
            return false;
        });
        //更新完成
        //丢弃该事务，被其他事务请求时顶掉，wait-die
        request_queue->upgrading_ = false;
        if(txn->get_state() == TransactionState::ABORTED){
            throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
        }
        it->granted_ = true;
    }

    if(exist_same_tid)return need_upgrade;

    LockRequest lock_request(txn->get_transaction_id(), lock_mode);
    auto lock_request_iter = request_queue->request_queue_.insert(request_queue->request_queue_.end(),lock_request);
    request_queue->cv_.wait(q_lock, [&request_queue, &lock_request_iter, &txn, this]{
        // wait die
        if(txn->get_state()==TransactionState::ABORTED){
            return true;
        }
        txn_id_t occupied_id = txn->get_transaction_id();
        if(checkLockToQueueCompatible(request_queue, &(*lock_request_iter), occupied_id)) {
            return true;
        }

        if(occupied_id < txn->get_transaction_id()){
            txn->set_state(TransactionState::ABORTED);
            //回收txn，aborted
            //request_queue->request_queue_.erase(lock_request_iter);
            return true;
        }
        return false;
    });
    //冲突已经解决
    if(txn->get_state() == TransactionState::ABORTED){
        request_queue->request_queue_.erase(lock_request_iter);
        throw TransactionAbortException (txn_id, AbortReason::DEADLOCK_PREVENTION);
    }

    lock_request_iter->granted_ = true;
    //request_queue->request_queue_.emplace_back(lock_request);
    return need_upgrade;
}

