/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "transaction_manager.h"
#include "record/rm_file_handle.h"

std::unordered_map<txn_id_t, Transaction *> TransactionManager::txn_map = {};
std::atomic<bool> enable_logging = false;
//注意点：这里涉及的操作，也要遵循WAL原则
/**
 * @description: 事务的开始方法
 * @return {Transaction*} 开始事务的指针
 * @param {Transaction*} txn 事务指针，空指针代表需要创建新事务，否则开始已有事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
Transaction * TransactionManager::begin(Transaction* txn, LogManager* log_manager) {
    // Todo:
    // 1. 判断传入事务参数是否为空指针
    // 2. 如果为空指针，创建新事务
    // 3. 把开始事务加入到全局事务表中
    // 4. 返回当前事务指针
    std::unique_lock<std::mutex> l(latch_);
    if(txn == nullptr){
        //this->setNextTid(log_manager->getLastTid() + 1);
        txn = new Transaction(next_txn_id_++);
        //printf("%d",int(next_txn_id_));
    }
    txn_map[txn->get_transaction_id()] = txn;

    if(enable_logging) {
        auto begin_log = new BeginLogRecord(txn->get_transaction_id());
        begin_log->prev_lsn_ = txn->get_prev_lsn();
        auto lsn = log_manager->add_log_to_buffer(begin_log);
        txn->set_prev_lsn(lsn);
    }
    return txn;
}

/**
 * @description: 事务的提交方法
 * @param {Transaction*} txn 需要提交的事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
void TransactionManager::commit(Transaction* txn, LogManager* log_manager) {
    // Todo:
    // 1. 如果存在未提交的写操作，提交所有的写操作
    // 2. 释放所有锁
    // 3. 释放事务相关资源，eg.锁集
    // 4. 把事务日志刷入磁盘中
    // 5. 更新事务状态
    ReleaseLock(txn);
    txn->get_write_set()->clear();

    if(enable_logging){
        auto commit_log = new CommitLogRecord(txn->get_transaction_id(),txn->get_prev_lsn());
        commit_log->prev_lsn_ = txn->get_prev_lsn();
        auto lsn = log_manager->add_log_to_buffer(commit_log);
        txn->set_prev_lsn(lsn);
        log_manager->flush_log_to_disk();
    }
    //log_manager->flush_log_to_disk();
    txn->set_state(TransactionState::COMMITTED);//所有log已经刷入
    std::unique_lock<std::mutex> l (latch_);
//    //全局事务表更新
    txn_map.erase(txn->get_transaction_id());
}

/**
 * @description: 事务的终止（回滚）方法
 * @param {Transaction *} txn 需要回滚的事务
 * @param {LogManager} *log_manager 日志管理器指针
 */
void TransactionManager::abort(Transaction * txn, LogManager *log_manager) {
    // Todo:
    // 1. 回滚所有写操作
    auto write_set = txn->get_write_set();
    Context ctx(lock_manager_,log_manager,txn);


    //针对表记录的写和索引的写
    while(!write_set ->empty()){
        auto &write_record_meta = write_set->back();
        auto &file_hl = sm_manager_->fhs_.at(write_record_meta->GetTableName());
        auto indexes = sm_manager_->db_.get_table(write_record_meta->GetTableName()).indexes;

        //先写完对应的回滚日志，再修改相应的数据
        switch (write_record_meta->GetWriteType()) {
            case WType::INSERT_TUPLE:{
                //auto rec = file_hl->get_record(write_record_meta->GetRid(), nullptr);//因为对于INSERRT，write_record.data为空
                //得到值后，需要unpin操作，这里放在最后进行
                if(enable_logging) {
                    auto d_log = new DeleteLogRecord(txn->get_transaction_id(), txn->get_prev_lsn(),
                                                     write_record_meta->GetRecord(),
                                                     write_record_meta->GetRid(), write_record_meta->GetTableName());
                    auto lsn = log_manager->add_log_to_buffer(d_log);
                    txn->set_prev_lsn(lsn);
                    Page* page = sm_manager_->get_bpm()->fetch_page({file_hl->GetFd(),write_record_meta->GetRid().page_no});
                    page->set_page_lsn(lsn);
                    sm_manager_->get_bpm()->unpin_page({file_hl->GetFd(),write_record_meta->GetRid().page_no},true);
                }
                //对于不同索引,从rec中提取key
                for(auto &index : indexes) {
                    //获取索引文件的handle
                    auto ih = sm_manager_->ihs_.at(
                            sm_manager_->get_ix_manager()->get_index_name(write_record_meta->GetTableName(),
                                                                          index.cols)).get();
                    char *key = new char[index.col_tot_len];
                    int offset = 0;
                    for (size_t j = 0; j < (size_t) index.col_num; ++j) {
                        memcpy(key + offset, write_record_meta->GetRecord().data + index.cols[j].offset, index.cols[j].len);
                        offset += index.cols[j].len;
                    }
                    ih->delete_entry(key, txn);
                }
                file_hl->delete_record(write_record_meta->GetRid(), nullptr);
                break;
            }
            case WType::UPDATE_TUPLE:{
                auto new_rec = file_hl->get_record(write_record_meta->GetRid(), nullptr);
                //writeRecord记录了被删除的value
                auto old_rec = write_record_meta->GetRecord();
                if(enable_logging) {
                    auto update_log = new UpdateLogRecord(txn->get_transaction_id(), txn->get_prev_lsn(), *new_rec,
                                                          write_record_meta->GetRecord(),
                                                          write_record_meta->GetRid(), write_record_meta->GetTableName());
                    //txn->set_prev_lsn(log_manager->add_log_to_buffer(update_log));
                    auto lsn = log_manager->add_log_to_buffer(update_log);
                    txn->set_prev_lsn(lsn);
                    Page* page = sm_manager_->get_bpm()->fetch_page({file_hl->GetFd(),write_record_meta->GetRid().page_no});
                    page->set_page_lsn(lsn);
                    sm_manager_->get_bpm()->unpin_page({file_hl->GetFd(),write_record_meta->GetRid().page_no}, true);
                }

                for(auto &index : indexes) {
                    //获取索引文件的handle
                    auto ih = sm_manager_->ihs_.at(
                            sm_manager_->get_ix_manager()->get_index_name(write_record_meta->GetTableName(),
                                                                          index.cols)).get();
                    char *new_key = new char[index.col_tot_len];
                    char *old_key = new char[index.col_tot_len];
                    int offset = 0;
                    for (size_t j = 0; j < (size_t) index.col_num; ++j) {
                        //
                        memcpy(new_key + offset, new_rec->data + index.cols[j].offset, index.cols[j].len);
                        memcpy(old_key + offset, old_rec.data + index.cols[j].offset,index.cols[j].len);
                        offset += index.cols[j].len;
                    }
                    ih->delete_entry(new_key,txn);
                    ih->insert_entry(old_key,write_record_meta->GetRid(),txn);
                }
                file_hl->update_record(write_record_meta->GetRid(),write_record_meta->GetRecord().data, nullptr);

                break;
            }
            case WType::DELETE_TUPLE:{
                if(enable_logging) {
                    auto insert_log = new InsertLogRecord(txn->get_transaction_id(), txn->get_prev_lsn(),
                                                          write_record_meta->GetRecord(),
                                                          write_record_meta->GetRid(), write_record_meta->GetTableName());

                    auto lsn = log_manager->add_log_to_buffer(insert_log);
                    txn->set_prev_lsn(lsn);
                    Page* page = sm_manager_->get_bpm()->fetch_page({file_hl->GetFd(),write_record_meta->GetRid().page_no});
                    page->set_page_lsn(lsn);
                    sm_manager_->get_bpm()->unpin_page({file_hl->GetFd(),write_record_meta->GetRid().page_no}, true);
                }

                //更新索引组信息
                for(auto &index : indexes) {
                    //获取索引文件的handle
                    auto ih = sm_manager_->ihs_.at(
                            sm_manager_->get_ix_manager()->get_index_name(write_record_meta->GetTableName(),
                                                                          index.cols)).get();
                    char *key = new char[index.col_tot_len];
                    int offset = 0;
                    for (size_t j = 0; j < (size_t) index.col_num; ++j) {
                        memcpy(key + offset, write_record_meta->GetRecord().data + index.cols[j].offset, index.cols[j].len);
                        offset += index.cols[j].len;
                    }
                    ih->insert_entry(key,write_record_meta->GetRid(),txn);
                }
                //delete后不会自动整理slot的空闲碎片
                file_hl->insert_record(write_record_meta->GetRid(), write_record_meta->GetRecord().data);
                break;
            }
        }


        write_set->pop_back();
        delete write_record_meta;
    }
    //释放记录写的表
    write_set->clear();

    // 2. 释放所有锁
    ReleaseLock(txn);
    // 3. 清空事务相关资源，eg.锁集

    // 4. 把事务日志刷入磁盘中
    if(enable_logging) {
        auto abort_log = new AbortLogRecord(txn->get_transaction_id(), txn->get_prev_lsn());
        auto lsn = log_manager->add_log_to_buffer(abort_log);
        txn->set_prev_lsn(lsn);
        log_manager->flush_log_to_disk();
    }
    // 5. 更新事务状态
    txn->set_state(TransactionState::ABORTED);

    std::unique_lock<std::mutex> l(latch_);
    txn_map.erase(txn->get_transaction_id());
}

/**
* helper function
*/
void TransactionManager::ReleaseLock(Transaction *txn) {
    //std::unique_lock<std::mutex> l(latch_);
    auto lock_set = txn->get_lock_set();
    for(auto  item : *lock_set){
        lock_manager_->unlock(txn,item);
    }
    lock_set->clear();
}
